#==============================================================================
# package      DE_EPAGES::Core::API::Html
# description  Methods for handling HTML tags
# state        public
#==============================================================================
package DE_EPAGES::Core::API::Html;
use base qw( Exporter );

use strict;
use HTML::Entities qw(decode_entities);
use HTML::TreeBuilder;
use URI;

use DE_EPAGES::Core::API::Constants;
use DE_EPAGES::Core::API::TestParameter qw (TestString TestUrl);
use DE_EPAGES::Core::API::Url qw(Relative2Absolute);

our @EXPORT_OK = qw(
    BalanceHTMLTags
    CloseOpenHTMLTags
    RemoveHtmlTags
    RemoveScriptTags
    RemoveSuspiciousChars
    ReplaceURIWithURL
);

#========================================================================================
# function     RemoveSuspiciousChars
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $SaveString = RemoveSuspiciousChars($StringWithHTML);
#----------------------------------------------------------------------------------------
# description  Removes all encodings for < and >
#----------------------------------------------------------------------------------------
# input        $StringWithHTML | string | string
# return       $SaveString | string without suspicious chars | string
#========================================================================================
sub RemoveSuspiciousChars {
    my ($Value) = @_;

    my @Encodings = (@{(ENCODINGS_LT)}, @{(ENCODINGS_GT)});
    my $Regex = join '|', @Encodings;
    my $RegexWithSemicolons = join ';|', @Encodings;

    # loop to elimitate nested encodings
    $Value =~ s/$RegexWithSemicolons;|$Regex//gi while ($Value =~ /$RegexWithSemicolons;|$Regex/i);

    # remove trailing whitespace
    $Value =~ s/^\s+//;
    $Value =~ s/\s+$//;

    return $Value;
}

#========================================================================================
# function     RemoveHtmlTags
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $NoHtml = RemoveHtmlTags($HTML);
#----------------------------------------------------------------------------------------
# description  Removes all html tags and consolidates spaces.
#----------------------------------------------------------------------------------------
# input        $HTML | html string | string
# return       $NoHtml | string without html tags | string
#========================================================================================
sub RemoveHtmlTags {
    local $_ = shift;

    s/[\x{0000}-\x{0008}\x{000B}\x{000C}\x{000E}-\x{001F}\x{007F}]//g; #delete all non printable characters
    s/\r|\n/ /g;                     # remove newlines
    s/<style\b.*?<\/style>/ /igs;    # remove style tags
    s/<script\b.*?<\/script>/ /igs;  # remove script tags
    s/\s+/ /g;       # remove multiple spaces
    s/<(?:br|hr)\s*\/?>/\n/igx;     #replace hr br with newline
    s/<\/(?:p|div|h\d|hr)>/\n/igx;    # replace block elements with newline
    s/<.*?>//gs;  # remove HTML tags
    s/\ +/ /g;       # consolidate multiple spaces between removed tags
    s/^\ //gm;       # spaces at the beginning of a line
    s/\ $//gm;       # spaces at the end of a line
    s/^\s+//;        # remove leading spaces
    s/\s+$//;        # remove trailing spaces
    s/&nbsp;/ /g;    # replace &nbsp; (decode_entities would replace with a special Unicode space)
    decode_entities($_);

    return $_;
}

#========================================================================================
# function     RemoveScriptTags
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $$NoScriptTagHtml = RemoveScriptTags($HTML);
#----------------------------------------------------------------------------------------
# description  Removes all script html tags
#----------------------------------------------------------------------------------------
# input        $HTML | html string | string
# return       $NoScriptTagHTML | string without script html tags | string
#========================================================================================
sub RemoveScriptTags {
    local $_ = shift // '';

    s/<script\b.*?<\/script>/ /igs;  # remove script tags

    return $_;
}

#========================================================================================
# function     BalanceHTMLTags
# state        private
#----------------------------------------------------------------------------------------
# syntax       $NewHtmlCode = BalanceHTMLTags($HtmlCode,[$hFlags]);
#               $hFlags = { 'getCompleteDocument' => 1 };
#----------------------------------------------------------------------------------------
# example      print BalanceHTMLTags('<p><span class=test>Hallo!');
#                          -> output: '<p></p><span class="test">Hallo!</span>'
#               print BalanceHTMLTags('<p><span class=test>Hallo!',{'getCompleteDocument'=>1});
#                          -> output: '<html><head></head><body><p></p><span class="test">Hallo!</span></body></html>'
#----------------------------------------------------------------------------------------
# description  Close any open left html tags to get a xhtml valid tag tree. Applies to
#               XHTML! Old html4 left-open tags are NOT allowed (like <P><UL> in html4).
#----------------------------------------------------------------------------------------
# input        $HtmlCode | html source code text | string
# input        $hFlags | getCompleteDocument - set this true if the function
#                         shall return a complete html document | ref.hash
# return       $NewHtmlCode | html source code with balanced html tags (xhtml) | string
#========================================================================================
sub BalanceHTMLTags($;$) {
    my ($HtmlCode,$hFlags) = @_;

    my $Tree = HTML::TreeBuilder->new;
    $Tree->store_comments(1);
    $Tree->p_strict(1);
    $Tree->parse_content($HtmlCode);

    my $NewHtmlCode = '';
    if($hFlags->{'getCompleteDocument'}){
        $NewHtmlCode = $Tree->as_HTML(undef, undef, {});
    } else {
        #Tree contains <html>, <head> and <body> nodes which are not needed
        #->get content of <head> node -> relevant for <script> and <style> snippets
        my @HeadContent = $Tree->look_down('_tag', 'head')->content_list;
        foreach my $ContentObject (@HeadContent){
            #press content objects to plain text/html, add missing html syntax (->as_HTML)
            $NewHtmlCode .= $ContentObject->as_HTML(undef, undef, {}) if ref $ContentObject eq 'HTML::Element';
            $NewHtmlCode .= $ContentObject if ref $ContentObject eq '';
        }
        #->get content of <body> node
        my @BodyContent = $Tree->look_down('_tag', 'body')->content_list;
        foreach my $ContentObject (@BodyContent){
            #press content objects to plain text/html, add missing html syntax (->as_HTML)
            $NewHtmlCode .= $ContentObject->as_HTML(undef, undef, {}) if ref $ContentObject eq 'HTML::Element';
            $NewHtmlCode .= $ContentObject if ref $ContentObject eq '';
        }
    }
    $Tree->delete;
    $NewHtmlCode =~ s/\n//g;
    return $NewHtmlCode;
}

#========================================================================================
# function     CloseOpenHTMLTags
# state        private
#----------------------------------------------------------------------------------------
# syntax       $NewHtmlCode = CloseOpenHTMLTags($HtmlCode);
#----------------------------------------------------------------------------------------
# example      print CloseOpenHTMLTags('<p><span class="test">Hallo!');
#                            -> output: '<p><span class="test">Hallo!</span></p>'
#----------------------------------------------------------------------------------------
# description  Close any open left html tags (xhtml). Does not correct loose quotations
#               nor DOM node attributes. Ignores script tag content.
#----------------------------------------------------------------------------------------
# input        $HtmlCode | html source code text | string
# return       $NewHtmlCode | html source code with balanced html tags (xhtml) | string
#========================================================================================
sub CloseOpenHTMLTags ($) {
    my ($SourceCode) = @_;

    my @TagStack;
    my $StackSize = 0;
    my $TagQueue = '';
    my $NewText = '';

    my @SingleTags = qw(br hr img input);      # self-closing tags like <img .. />
    my @NestableTags = qw(blockquote div span ul ol);   # Tags that can be immediately nested within themselves

    my $Text = $SourceCode;
    # fix weird comments like '< !--'
    $Text =~ s/<\s!--/<!\x20{4}--/g;

    # while there're html tags
    while($Text =~ /(<(\/?\w*)\s*([^>]*)>)/xg) {  # match html tags
        my $RegEx0 = $1;    # entire string
        my $RegEx1 = $2;    # tag with < >
        my $RegEx2 = $3;    # tag

        $NewText .= $TagQueue;

        # offset and size
        my $i = index($Text, $RegEx0);
        my $l = length($RegEx0);

        # clear the shifter
        $TagQueue = '';
        my $Tag;

        # Pop or Push
        if($RegEx1 and $RegEx1 =~ /^\//) {     # found html end tag (</tag>)
            $Tag = lc(substr($RegEx1,1));
            if($StackSize <= 0) {
                $Tag = '' ;      # if too many closing tags
            } elsif ($TagStack[$StackSize - 1] eq $Tag) { # found closing tag
                $Tag = '</' . $Tag . '>'; # close Tag
                # Pop
                pop (@TagStack);
                $StackSize--;
            } else { # closing tag not at top, search for it
                if("@TagStack" =~ /script/) { # if script tag use found closing tag - don't try to correct html inside javascript!
                    $TagQueue .= '</' . $Tag . '>';
                } else {
                    for (my $j=$StackSize-1;$j>=0;$j--) {
                        if ($TagStack[$j] eq $Tag) {
                        # add tag to tagqueue
                            for (my $k=$StackSize-1;$k>=$j;$k--){
                                $TagQueue .= '</' . pop(@TagStack) . '>';
                                $StackSize--;
                            }
                            last;
                        }
                    }
                }
                $Tag = '';
            }
        } else {    # found html begin tag (<tag>)
            $Tag = lc($RegEx1);

            # don't try to correct html inside javascript, e.g. selectors
            local $" = ','; #"
            next if  "@TagStack" =~ /script/;
            # If self-closing or '', don't do anything.
            if((substr($RegEx2,-1) eq '/') || ($Tag eq '')) {
                # noop
            } elsif ( grep { $_ eq $Tag } @SingleTags ) {  # elseif it's a self-closing entity tag but it doesn't close itself, do so
                $RegEx2 .= '/';
            } else {    # push tag onto the stack
                # If the top of the stack is the same as the tag we want to push, close previous tag
                if (($StackSize > 0) && !(grep { $_ eq $Tag } @NestableTags) && ($TagStack[$StackSize - 1] eq $Tag)) {
                    $TagQueue = '</' . pop(@TagStack) . '>';
                    $StackSize--;
                }

                #print "tag: $RegEx1; parent tag: ". $TagStack[$StackSize - 1] ."\n" if defined $TagStack[$StackSize - 1];

                push(@TagStack, $Tag);
                $StackSize = scalar(@TagStack);     # refresh stack size
            }

            # attributes
            my $Attributes = $RegEx2;               # like: color="#FFFFFF" style="..."
            if($Attributes) {
                $Attributes = ' '.$Attributes;
            }
            $Tag = '<'.$Tag.$Attributes.'>';        # like: <p color="#FFFFFF" style="...">
            if ($TagQueue) {                        # If already queuing a close tag, then put this tag on, too
                $TagQueue .= $Tag;
                $Tag = '';
            }
        }

        $NewText .= substr($Text,0,$i) . $Tag;      # append
        $Text = substr($Text,$i+$l);                # update remaining text
    }

    # Clear Tag Queue
    $NewText .= $TagQueue;

    # Add Remaining text
    $NewText .= $Text;

    # Empty Stack
    while(my $x = pop(@TagStack)) {
        $NewText .= '</' . $x . '>';                # Add remaining tags to close
    }

    # rebuild HTML comments
    $NewText =~ s/<\s!--/<!--/g;
    $NewText =~ s/<\s{4}!--/<\x20!--/g;

    return $NewText;
}

#========================================================================================
# function     ReplaceURIWithURL
# state        private
#----------------------------------------------------------------------------------------
# syntax       $NewHtmlCode = ReplaceURIWithURL($HtmlCode, $BaseURL);
#----------------------------------------------------------------------------------------
# example      ReplaceURIWithURL('<a href="test.html">Hallo!</a>', 'www.example.com');
#               > output: '<p><span class="test"><a href="www.example.com/test.html">Hallo!</span></p>'
#----------------------------------------------------------------------------------------
# description  Replaces uri in tag attributes with complete url.
#----------------------------------------------------------------------------------------
# input        $HtmlCode | html source code text | string
# input        $BaseURL | base url | string
# input        $aTags | list of tags to check (optional) | ref.array
# return       $NewHtmlCode | html source code with converted uris | string
#========================================================================================
sub ReplaceURIWithURL ($$;$) {
    my ($HtmlCode, $BaseURL, $aTags) = @_;
    TestString('HTMLCode', $HtmlCode);
    TestUrl('BaseURL', $BaseURL);

    # Build DOM from string.
    my $Tree = HTML::TreeBuilder->new;
    $Tree->no_space_compacting(1);
    $Tree->parse($HtmlCode);

    my %TagHash = %{(TAG_URI_ATTRIBUTE)};
    # Set if only some special tags should check.
    if (defined $aTags and scalar @$aTags > 0) {
        %TagHash = map { $_ => TAG_URI_ATTRIBUTE->{$_} } grep { defined TAG_URI_ATTRIBUTE->{$_} } @$aTags;
    }

    while (my ($SearchTag, $Attribute) = each %TagHash) {
        my @Tags = $Tree->look_down('_tag', $SearchTag);
        foreach my $Tag (@Tags) {
            my $URI = $Tag->attr($Attribute);
            # Build absolute path only if there is an uri and the uri is not an anchor (starting with #).
            if (defined $URI and $URI ne '' and $URI =~ /\S/ and $URI !~ /^#/) {
                my $URL = Relative2Absolute($URI, $BaseURL);
                $Tag->attr($Attribute, $URL);
            }
        }
    }

    return $Tree->guts()->as_HTML('<>&"\'');
}

1;
