#========================================================================================
# package      DE_EPAGES::DHLIntraship::t::RemoteGetPackstationsByAddress
# base         Test::Class
# state        private
#----------------------------------------------------------------------------------------
# description  Tests for getPackstationByAddress SOAP call (integration tests)
#========================================================================================
package DE_EPAGES::DHLIntraship::t::RemoteGetPackstationsByAddress;

use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Database::API::Connection qw ( RunOnStore GetTestSite );
use DE_EPAGES::Object::API::Factory qw(LoadObjectByPath ExistsObjectByPath LoadRootObject);
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw ( ImportObjects );

use DE_EPAGES::DHLIntraship::API::SOAP::PackstationRequest;
use DE_EPAGES::DHLIntraship::API::SOAP::PackstationResponse;
use DE_EPAGES::DHLIntraship::API::SOAP::WebServiceClient;

use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::MockifyServlet;

use Test::More;
use Test::Exception;
use Test::Deep;

use strict;

use constant ROOTOBJECT => 'Test';
use constant ROOTPATH => '/' . ROOTOBJECT;
use constant SHOPPATH => ROOTPATH . '/TestShopDHLVersenden';

use constant TEST_DE_ADDRESS_STREET_NAME => 'Pilatuspool';
use constant TEST_DE_ADDRESS_HOUSE_NUMBER => '2';
use constant TEST_DE_ADDRESS_ZIP_CODE => '20355';
use constant TEST_DE_ADDRESS_CITY => 'Hamburg';
use constant TEST_NON_DE_ADDRESS_STREET_NAME => 'Diagonal';
use constant TEST_NON_EXISTENT_ADDRESS_STREET_NAME => 'NonExistenStreeName';
use constant TEST_NON_DE_ADDRESS_HOUSE_NUMBER => '435';
use constant TEST_NON_DE_ADDRESS_ZIP_CODE => '08036';
use constant TEST_NON_DE_ADDRESS_CITY => 'Barcelona';


sub Init {

    return if ExistsObjectByPath(ROOTPATH) && $ENV{'SKIP_CLEANUP'};

    Cleanup();

    my $System = LoadRootObject();
    $System->insertChild( 'Object', { Alias => ROOTOBJECT } );
    note 'importing shop ' . SHOPPATH;
    ImportObjects('import.xml', ROOTPATH);

    return;
}

sub Cleanup {

    return if $ENV{'SKIP_CLEANUP'};

    my $System = LoadRootObject();
    if( $System->existsChild(ROOTOBJECT) ) {
        note 'deleting shop ' . SHOPPATH;
        $System->child(ROOTOBJECT)->delete();
    }

    return;
}

sub test {
    my $self = shift;

    $self->packstation_request_positive_test();
    $self->packstation_request_negative_non_german_address_test();
    $self->packstation_request_negative_non_existent_address_test();
}

#========================================================================================
# Test methods
#========================================================================================
sub test_xml_import {
    my $subTestName = (caller(0))[3];

    my $testImportOK = 0;

    my $Shop = LoadObjectByPath(SHOPPATH);
    $testImportOK = isa_ok($Shop, 'DE_EPAGES::Shop::API::Object::Shop', "$subTestName - test shop found");

    my $ShippingFulfillmentMethod = $Shop->path('ShippingFulfillmentMethods/DHLIntraship');
    $testImportOK = isa_ok($ShippingFulfillmentMethod,
            'DE_EPAGES::DHLIntraship::API::Object::ShippingFulfillmentMethodDHLIntraship',
            "$subTestName - test ShippingFulfillmentMethodDHLIntraship found")
                        && $testImportOK;

    return $testImportOK;
}

sub packstation_request_positive_test {
    my $self = shift;
    my $subTestName = (caller(0))[3];

    my $hGermanAddress = {
            'Street' => TEST_DE_ADDRESS_STREET_NAME,
            'StreetNo' => TEST_DE_ADDRESS_HOUSE_NUMBER,
            'ZipCode' => TEST_DE_ADDRESS_ZIP_CODE,
            'City' => TEST_DE_ADDRESS_CITY,
    };
    my $aPackstations = $self->_getPackstationsFromService($hGermanAddress);

    cmp_ok(scalar @{$aPackstations}, '>', 0, "$subTestName - one or more packstations returned");

    my $hPackstation = $aPackstations->[0];
    cmp_deeply($hPackstation,
            {
                'packstationId' => ignore(),
                'district' => ignore(),
                'longitude' => ignore(),
                'distance' => ignore(),
                'name' => ignore(),
                'zipcode' => ignore(),
                'city' => ignore(),
                'latitude' => ignore(),
                'address' => ignore()
            },
            "$subTestName - packstation hash structure"
    );
}

sub packstation_request_negative_non_german_address_test {
    my $self = shift;
    my $subTestName = (caller(0))[3];

    my $hNonGermanAddress = {
            'Street' => TEST_NON_DE_ADDRESS_STREET_NAME,
            'StreetNo' => TEST_NON_DE_ADDRESS_HOUSE_NUMBER,
            'ZipCode' => TEST_NON_DE_ADDRESS_ZIP_CODE,
            'City' => TEST_NON_DE_ADDRESS_CITY,
    };

    my $aPackstations = $self->_getPackstationsFromService($hNonGermanAddress);

    cmp_ok(scalar @{$aPackstations}, '==', 0, "$subTestName - zero packstations returned (over a non German address)");
}

sub packstation_request_negative_non_existent_address_test {
    my $self = shift;
    my $subTestName = (caller(0))[3];

    my $hGermanButNonExistentStreetAddress = {
            'Street' => TEST_NON_EXISTENT_ADDRESS_STREET_NAME,
            'StreetNo' => TEST_DE_ADDRESS_HOUSE_NUMBER,
            'ZipCode' => TEST_DE_ADDRESS_ZIP_CODE,
            'City' => TEST_DE_ADDRESS_CITY,
    };

    my $aPackstations = $self->_getPackstationsFromService($hGermanButNonExistentStreetAddress);

    cmp_ok(scalar @{$aPackstations}, '==', 0,
            "$subTestName - zero packstations returned (over a non existent street address)");
}

#========================================================================================
# Private methods
#========================================================================================
sub _getPackstationsFromService {
    my $self = shift;
    my ($hAddress) = @_;

    my $Shop = LoadObjectByPath(SHOPPATH);
    my $ShippingFulfillmentMethod = $Shop->path('ShippingFulfillmentMethods/DHLIntraship');

    my $MockifyServlet = $self->_createServletWithAddressParams($ShippingFulfillmentMethod, $hAddress);

    my $DHLWSClient = DE_EPAGES::DHLIntraship::API::SOAP::WebServiceClient->new(
        DE_EPAGES::DHLIntraship::API::SOAP::PackstationRequest->new(),
        DE_EPAGES::DHLIntraship::API::SOAP::PackstationResponse->new(),
        $ShippingFulfillmentMethod->get('IsInTestMode')
    );

    return $DHLWSClient->getPackstationsByAddress($MockifyServlet);
}

#========================================================================================
# Mocking methods
#========================================================================================
sub _createServletWithAddressParams {
    my $self = shift;
    my ($ShippingFulfillmentMethod, $hAddressParams) = @_;

    my $MockifyServlet = DE_EPAGES::Test::Mock::MockifyServlet->new( $ShippingFulfillmentMethod );

    $MockifyServlet->addParams($hAddressParams);

    return $MockifyServlet->getMockObject();
}


RunScript(
    Sub => sub {
        RunOnStore( 'Store' => GetTestSite(), Sub => sub {
            Init();
            unless ( test_xml_import() ) {
                done_testing();
                exit;
            }

            SKIP: {
                skip 'external services tests' if $ENV{'SKIP_EXTERNAL_TESTS'};
                my $self = __PACKAGE__;
                $self->test();
            }

            done_testing();
            Cleanup();
        });
    }
);

1;