use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Error qw( GetError ExistsError );
use DE_EPAGES::Core::API::Log qw( LogError );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Domain qw ( GetUnitySubDomain );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadObjectByPath );
use DE_EPAGES::Shop::API::System qw( GetUnityShopIDs );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );


#========================================================================================
# function         FixUnityLogoUrl
# state            public
#----------------------------------------------------------------------------------------
# syntax           FixUnityLogoUrl();
#----------------------------------------------------------------------------------------
# description      Fixes logo path for unity shop, if it contains '/WebRoot/' or multiple
#                   '&shop=' parameters.
#========================================================================================
sub FixUnityLogoUrl {
    my ($ShopAlias) = @_;

    if ( defined $ShopAlias ) {
        my $Shop = LoadObjectByPath('/Shops/'.$ShopAlias);
        if ($Shop->isUnity()) {
            # Unity shop
            _fixUnityLogoUrl($Shop);
        } else {
            # non Unity shop
           print 'Shop ' . $ShopAlias . " is not a Unity shop.\n";
        }
    } else {
        print "Reading all Unity shops ...\n";
        my $aShopIDs = GetUnityShopIDs();
        foreach my $ShopID (@$aShopIDs) {
            eval {
                my $Shop = LoadObject($ShopID);
                _fixUnityLogoUrl($Shop);
            };
            if (ExistsError()) {
                print("Patch failed in shop " . $ShopID . " with error \"" . GetError() . "\".\n");
                warn("Patch failed in shop " . $ShopID . " with error \"" . GetError() . "\".\n");
                next;
            }
        }
    }

    eval { SynchronizeCache(); };
    LogError(GetError()) if ExistsError();

    return;
}


sub _fixUnityLogoUrl {
    my ($Shop) = @_;

    my $ShopAlias = $Shop->get('Alias');

    my $Logo = $Shop->get('Logo');
    my $OrigLogo = $Shop->get('Logo');
    return unless defined($Logo);

    my $hPatchValues = {};

    # Removing appearance of mutiple &shop= parameters
    if ($Logo =~ /(&shop=.*){2,}/) {
        print "Logo URL contains multiple '&shop=' parameter\n";
        $Logo =~ s/&shop=.*//g;
        $Logo .= '&shop=' . GetUnitySubDomain($Shop->get('WebServerScriptNamePart'));
        $hPatchValues->{'Logo'} = $Logo;
    }

    # Removing WebRoot path if path also contains /storage/images
    if ($Logo =~ /(WebRoot\/.*\/)\/storage\/images/) {
        print "Logo URL contains unnecessary WebRoot path\n";
        $Logo =~ s/WebRoot\/.*\/\///;
        $hPatchValues->{'Logo'} = $Logo;
    }

    if ($Logo ne $OrigLogo) {
        print "> ", $ShopAlias, "\n";
        print ">> before: ", $OrigLogo, "\n";
        print ">> after: ", $Logo, "\n";
    }

    $Shop->set($hPatchValues);
}

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $ShopAlias;
    my $Help;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'shop=s'      => \$ShopAlias,
    );

    usage() if $Help;

    unless ( defined $StoreName ) {
        print STDERR "Missing parameter -storename.\n\n";
        usage();
    }
    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub { FixUnityLogoUrl($ShopAlias); },
    );
    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

sub usage {
    print <<END_USAGE;
Description:
    Fixes logo path for unity shop, if it contains '/WebRoot/' or multiple
    '&shop=' parameters. Prints only changed logo urls.

    Bug was introduced, after running 'patchUnityLogoAndFavicon.pl'.

Usage:
    perl $0 [options] [flags]

options:
    -passwd     database user password                          ()
    -storename  name of store                                   ()
    -shop       alias                              (optional)   ()
flags:
    -help       show the command line options

Example:
    perl $0 -storename Store
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );