use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Error qw( GetError ExistsError );
use DE_EPAGES::Core::API::Log qw( LogError );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Domain qw ( GetUnitySubDomain );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadObjectByPath );
use DE_EPAGES::Shop::API::System qw( GetUnityShopIDs );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );


#========================================================================================
# function         PatchUnityLogoAndFavicon
# state            public
#----------------------------------------------------------------------------------------
# syntax           PatchUnityLogoAndFavicon();
#----------------------------------------------------------------------------------------
# description      Patches Logo, ShouldHideLogo and FavIcon.
#========================================================================================
sub PatchUnityLogoAndFavicon {
    my ($ShopAlias) = @_;

    if ( defined $ShopAlias ) {
        my $Shop = LoadObjectByPath('/Shops/'.$ShopAlias);
        if ($Shop->isUnity()) {
            # Unity shop
            _patchUnityLogoAndFavicon($Shop);
        } else {
            # non Unity shop
           print 'Shop ' . $ShopAlias . " is not a Unity shop.\n";
        }
    } else {
        print "Reading all Unity shops ...\n";
        my $aShopIDs = GetUnityShopIDs();
        foreach my $ShopID (@$aShopIDs) {
            eval {
                my $Shop = LoadObject($ShopID);
                _patchUnityLogoAndFavicon($Shop);
            };
            if (ExistsError()) {
                print("Patch failed in shop " . $ShopID . " with error \"" . GetError() . "\".\n");
                warn("Patch failed in shop " . $ShopID . " with error \"" . GetError() . "\".\n");
                next;
            }
        }
    }

    eval { SynchronizeCache(); };
    LogError(GetError()) if ExistsError();

    return;
}


sub _patchUnityLogoAndFavicon {
    my ($Shop) = @_;

    my $ShopAlias = $Shop->get('Alias');
    my $ShopContent;
    eval {
        $ShopContent = $Shop->sendUnityRequest('GET', '/shop');
    };
    if (ExistsError()) {
        print("Sending request to Unity failed in shop " . $ShopAlias. " with error \"" . GetError() . "\".\n");
        warn("Sending request to Unity failed in shop " . $ShopAlias . " with error \"" . GetError() . "\".\n");
        return;
    }

    my $WebServerScriptNamePart = $Shop->get('WebServerScriptNamePart');
    my $LogoUrl = $ShopContent->{'logoUrl'};
    return unless defined($LogoUrl);

    $LogoUrl = $LogoUrl . '&shop=' . GetUnitySubDomain($WebServerScriptNamePart) if ($LogoUrl !~ /&shop=/);

    my $ShouldHideLogo = $ShopContent->{'shouldHideLogo'} | 0; # convert true/false to 1/0
    my $FaviconUrl = $ShopContent->{'faviconUrl'};

    my $hPatchValues = {};
    # set Logo (also set UseUnityShopLogo, just to be sure)
    # if '/image\?remote=/' in logo path, then logo is saved in epages ui
    if ($Shop->get('Logo') ne $LogoUrl and $LogoUrl =~ /image\?remote=/) {
        $Shop->set({
            "UseUnityShopLogo" => 1
        });
        print "Patch Logo for ". $ShopAlias . " to " . $LogoUrl . " ...\n";
        $hPatchValues->{'Logo'} = $LogoUrl;
    }

    # set ShouldHideLogo
    if ($Shop->get('ShouldHideLogo') != $ShouldHideLogo) {
        print "Patch ShouldHideLogo for ". $ShopAlias . " to " . $ShouldHideLogo . " ...\n";
        $hPatchValues->{'ShouldHideLogo'} = $ShouldHideLogo;
    }

    # set FavIcon
    if ($Shop->get('FavIcon') ne $FaviconUrl and $FaviconUrl =~ /^\/storage\/images/gi ) {
        print "Patch FavIcon for ". $ShopAlias . " to " . $FaviconUrl . " ...\n";
        $hPatchValues->{'FavIcon'} = $FaviconUrl;
    }

    $Shop->set($hPatchValues);
}

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $ShopAlias;
    my $Help;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'shop=s'      => \$ShopAlias,
    );

    usage() if $Help;

    unless ( defined $StoreName ) {
        print STDERR "Missing parameter -storename.\n\n";
        usage();
    }
    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub { PatchUnityLogoAndFavicon($ShopAlias); },
    );
    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

sub usage {
    print <<END_USAGE;
Description:
    Patches Logo, ShouldHideLogo and FavIcon.

Usage:
    perl $0 [options] [flags]

options:
    -passwd     database user password                          ()
    -storename  name of store                                   ()
    -shop       alias                              (optional)   ()
flags:
    -help       show the command line options

Example:
    perl $0 -storename Store
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );