use Test::More tests => 65;
use strict;
use DE_EPAGES::Core::API::Error qw( ExistsError GetError );
use DE_EPAGES::Core::API::Script qw (RunScript);
use DE_EPAGES::Database::API::Connection qw( ConnectTestSite DisconnectSite );
use DE_EPAGES::Object::API::Table::Country qw ( GetInfoCountriesByCode2 );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject LoadObjectByPath ExistsObjectByPath );
use DE_EPAGES::Shop::API::Address qw ( CheckVATID UpdateVATID IsTaxModelNet IsPrivateVATID );

my $ROOTPATH = '/Test';

sub Main {
    ConnectTestSite();

    Init();

    my $DE = GetInfoCountriesByCode2('DE')->[0]->{'CountryID'};
    my $ES = GetInfoCountriesByCode2('ES')->[0]->{'CountryID'};
    my $PT = GetInfoCountriesByCode2('PT')->[0]->{'CountryID'};

    ok (!IsPrivateVATID('DE 282 947 700'), '\'DE 282 947 700\' is a business VATID');
    ok (!IsPrivateVATID('ES B64 817 133'), '\'ES B64 817 133\' is a business VATID');
    ok (!IsPrivateVATID('PT 502 816 481'), '\'PT 502 816 481\' is a business VATID');
    ok (IsPrivateVATID('PT 123 456 789'), '\'PT 123 456 789\' is a private VATID');

    ok (CheckVATID('DE 282 947 700', $DE) == 1, '\'DE 282 947 700\' is valid in DE');
    ok (CheckVATID('AT 123 456 789', $DE) == -1, '\'AT 123 456 789\' is not valid in DE');
    ok (CheckVATID('DE 123 456 789 012', $DE) == 0, '\'DE 123 456 789 012\' is not valid in DE');
    ok (CheckVATID('DE 123 456 789', $DE) == 0, '\'DE 123 456 789\' is not valid in DE');
    ok (CheckVATID('DE 282-947-700', $DE) == 1, '\'DE 282-947-700\' is valid in DE');

    my $IT = GetInfoCountriesByCode2('IT')->[0]->{'CountryID'};
    ok (CheckVATID('DE 123 456 789', $IT) == -1, '\'DE 123 456 789\' is not in IT');
    ok (CheckVATID('IT 123 456 789', $IT) == 0, '\'IT 123 456 789\' is not a valid VATID in IT');
    ok (CheckVATID('IT 12345678901', $IT) == 1, '\'IT 12345678901\' is a valid VATID in IT');

    ok (CheckVATID('DE 123 456 789', $PT) == -1, '\'DE 123 456 789\' is not in PT');
    ok (CheckVATID('PT 501 749 01', $PT) == 0, '\'PT 501 749 01\' is not a valid VATID in PT');
    ok (CheckVATID('PT 501 749 1337', $PT) == 0, '\'PT 501 749 1337\' is not a valid VATID in PT');
    ok (CheckVATID('PT 501 749 012', $PT) == 1, '\'PT 501 749 012\' is a valid VATID in PT');
    ok (CheckVATID('PT 505 062 127', $PT) == 1, '\'PT 505 062 127\' is a valid VATID in PT');

    my $AT = GetInfoCountriesByCode2('AT')->[0]->{'CountryID'};
    ok (CheckVATID('AT U 154 167 07', $AT) == 1, '\'AT U 154 167 07\' is valid in AT');
    ok (CheckVATID('DE 123 456 789', $AT) == -1, '\'DE 123 456 789\' is not in AT');
    ok (CheckVATID('AT 123 456 789', $AT) == 0, '\'AT 123 456 789\' is not valid in AT');

    my $CH = GetInfoCountriesByCode2('CH')->[0]->{'CountryID'};
    ok (CheckVATID('CH 123 456 789', $CH) == 1, '\'CH 123 456 789\' is valid in CH');
    ok (CheckVATID('XY 123 456 789', $CH) == 1, '\'XY 123 456 789\' is valid in CH');
    ok (CheckVATID('AT 123 456 789', $CH) == 1, '\'AT 123 456 789\' is valid in CH');
    ok (CheckVATID('DE 123 456 789', $CH) == 1, '\'DE 123 456 789\' is valid in CH');

    my $AU = GetInfoCountriesByCode2('AU')->[0]->{'CountryID'};
    ok (CheckVATID('XYZ', $AU) == 1, 'any VAT-ID is valid for non-EU countries');
    ok (CheckVATID('XYZ') == 1, 'any VAT-ID is valid if no country is specified');

    my $GR = GetInfoCountriesByCode2('GR')->[0]->{'CountryID'};
    ok (CheckVATID('123 456 789', $GR) == -1, '\'123 456 789\' is not valid in GR without countrycode');
    ok (CheckVATID('GR 123 456 789', $GR) == -1, '\'GR 123 456 789\' is not in GR');
    ok (CheckVATID('EL 094 468 339', $GR) == 1, '\'EL 094 468 339\' is valid in GR');

    my $MC = GetInfoCountriesByCode2('MC')->[0]->{'CountryID'};
    ok (CheckVATID('123 456 789', $MC) == -1, '\'123 456 789\' is not valid in MC without countrycode');
    ok (CheckVATID('MC 123 456 789', $MC) == -1, '\'MC 123 456 789\' is not in MC');
    ok (CheckVATID('FR 30000105154', $MC) == 1, '\'FR 30000105154\' is valid in MC');

    ok (CheckVATID('ES B64 817 133', $ES) == 1, '\'ES B64 817 133\' is a valid business VATID in ES');
    ok (!CheckVATID('ES 123 456 789', $ES) == 1, '\'ES 123 456 789\' is a valid private VATID in ES');

    ok (CheckVATID('PT 502 816 481', $PT) == 1, '\'PT 502 816 481\' is a valid business VATID in PT');
    ok (CheckVATID('PT 123 456 789', $PT) == 1, '\'PT 123 456 789\' is a valid private VATID in PT');

    eval { CheckVATID(undef, $DE) };
    ok( ExistsError(), 'CheckVATID - missing VATID');
    is( GetError()->code, 'PARAMETER_MISSING', 'CheckVATID - missing VATID - error code');

    is (UpdateVATID('DE 123 456 789', $DE), 'DE123456789', 'VATID with countrycode');
    is (UpdateVATID('EU 123 456 789', $DE), 'EU123456789', 'VATID with EU countrycode');
    is (UpdateVATID('123 456 789', $DE), 'DE123456789', 'VATID without countrycode for DE');
    is (UpdateVATID('123 456 789', $AT), 'AT123456789', 'VATID without countrycode for AT');
    is (UpdateVATID('123 456 789'), '123456789', 'VATID without country id');
    is (UpdateVATID('EL 123 456 789', $GR), 'EL123456789', 'VATID with countrycode for GR');
    is (UpdateVATID('123 456 789', $GR), 'EL123456789', 'VATID without countrycode for GR');
    is (UpdateVATID('GR 123 456 789', $GR), 'ELGR123456789', 'VATID without countrycode for GR');
    is (UpdateVATID('DE 123 456 789', $GR), 'DE123456789', 'VATID with wrong countrycode for GR');

    is (UpdateVATID('DE 123-456-789', $DE), 'DE123456789', 'VATID with character "-"');

    # test error conditions
    eval { UpdateVATID(undef, $DE) };
    ok( ExistsError(), 'UpdateVATID - missing VATID');
    is( GetError()->code, 'PARAMETER_MISSING', 'UpdateVATID - missing VATID - error code');

    # test VATIDs (for switching customer to gross or net)
    my $Address = LoadObjectByPath('/Test/Address');
    my $Shop = LoadObjectByPath('/Test/TestShop1');
    my $AR = GetInfoCountriesByCode2('AR')->[0]->{'CountryID'};

    my $ShopCountryID = $Shop->get('Address')->get('CountryID');
    ok( !defined $ShopCountryID, 'new shop has not selected a country yet' );
    $Shop->get('Address')->set({'CountryID' => $DE});
    is( $Shop->get('Address')->get('CountryID'), $DE, 'shop country by address: GB' );

    $Address->set({'CountryID'=> $ES , 'VATID'    => '', });
    ok( !IsTaxModelNet($Shop, $Address), '"ES 25719487K" is a private NIF --> gross');

    $Address->set({'CountryID'=> $ES , 'VATID'    => 'ES L25719487K', });
    ok( !IsTaxModelNet($Shop, $Address), '"ES L25719487K" is a private NIF --> gross');

    $Address->set({'CountryID'=> $ES , 'VATID'    => 'ES B64817133', });
    ok( IsTaxModelNet($Shop, $Address), '"ES B64817133" is a business NIF --> net');

    $Address->set({'CountryID'=> $DE , 'VATID'    => 'DE 282 947 700', });
    ok( IsTaxModelNet($Shop, $Address), '"DE 282 947 700" is a business VAT ID --> net');

    $Address->set({'CountryID'=> $AR , 'VATID'    => '', });
    ok( IsTaxModelNet($Shop, $Address), 'non-EU foreigner (private) --> net');

    $Address->set({'CountryID'=> $AR , 'VATID'    => 'AG12345', });
    ok( IsTaxModelNet($Shop, $Address), 'non-EU foreigner (business) --> net');

    $Address->set({'CountryID'=> $ES , 'VATID'    => '', });
    ok( !IsTaxModelNet($Shop, $Address), 'EU foreigner without VAT ID --> gross');

    $Address->set({'CountryID'=> $DE , 'VATID'    => '', });
    ok( !IsTaxModelNet($Shop, $Address), 'Inland customer without VAT ID --> gross');

    # non-EU shop: set shop address country to Argentina
    $Shop->get('Address')->set({'CountryID' => $AR});
    is( $Shop->get('Address')->get('CountryID'), $AR, 'shop country by address: AR' );

    $Address->set({'CountryID'=> $AR , 'VATID'    => '', });
    ok( !IsTaxModelNet($Shop, $Address), 'non-eu-shop: Inland customer without VAT ID --> gross');

    $Address->set({'CountryID'=> $AR , 'VATID'    => 'AG12345', });
    ok( !IsTaxModelNet($Shop, $Address), 'non-eu-shop: Inland customer with VAT ID --> gross'); # because we can't verify the VAT ID

    $Address->set({'CountryID'=> $ES , 'VATID'    => '', });
    ok( IsTaxModelNet($Shop, $Address), 'non-eu-shop: Foreigner without VAT ID --> net');

    $Address->set({'CountryID'=> $ES , 'VATID'    => 'ES B64817133', });
    ok( IsTaxModelNet($Shop, $Address), 'non-eu-shop: Foreigner with VAT ID --> net');

    Cleanup();

    DisconnectSite();
}

sub Cleanup {

    LoadObjectByPath('/Test')->delete if ExistsObjectByPath('/Test');
    return 1;
}

sub Init {
    Cleanup();

    my $Folder = LoadRootObject()->insertChild( 'Object', { Alias => 'Test' });
    $Folder->insertChild( 'Shop', { Alias => 'TestShop1' } );

    $Folder->insertChild( 'Address', {
        'Alias'   => 'Address',
        'Street'  => 'Leutragraben',
        'City'    => 'Jena',
        'Zipcode' => '07740',
        'Company' => 'epages',
    });
    return 1;
}


RunScript(Sub=>\&Main, Debug=>1);


#'CY' => '[0-9]{8}[A-Za-z]',
#'IE' => '[0-9][A-Za-z0-9\\+\\*][0-9]{5}[A-Za-z]',
#'GB' => '([0-9]{3} ?[0-9]{4} ?[0-9]{2}|[0-9]{3} ?[0-9]{4} ?[0-9]{2} ?[0-9]{3}|GD[0-9]{3}|HA[0-9]{3})',
#'NL' => '[0-9]{9}B[0-9]{2}',
#'AT' => 'U[0-9]{8}',
#'BE' => '0[0-9]{9}',
#'EL' => '[0-9]{9}',
#'EE' => '[0-9]{9}',
#'LV' => '[0-9]{11}',
#'DK' => '[0-9]{2} ?[0-9]{2} ?[0-9]{2} ?[0-9]{2}',
#'PT' => '[0-9]{9}',
#'FI' => '[0-9]{8}',
#'DE' => '[0-9]{9}',
#'SE' => '[0-9]{12}',
#'BG' => '[0-9]{9,10}',
#'LU' => '[0-9]{8}',
#'SI' => '[0-9]{8}',
#'MT' => '[0-9]{8}',
#'LT' => '([0-9]{9}|[0-9]{12})',
#'RO' => '[0-9]{2,10}',
#'IT' => '[0-9]{11}',
#'PL' => '[0-9]{10}',
#'CZ' => '[0-9]{8,10}',
#'ES' => '([A-Za-z0-9][0-9]{7}[A-Za-z0-9])',
#'SK' => '[0-9]{10}',
#'HU' => '[0-9]{8}',
#'FR' => '[A-Za-z0-9]{2} ?[0-9]{9}'