#========================================================================================
# package      DE_EPAGES::ShopMigration::API::ProductTypesCommon
# state        public
#----------------------------------------------------------------------------------------
# description common functions for exporting/importing product types
#========================================================================================
package DE_EPAGES::ShopMigration::API::ProductTypesCommon;
use base qw(Exporter);

use strict;
use feature 'switch';
use DE_EPAGES::Core::API::PerlTools qw(ElementOf);
use DE_EPAGES::Core::API::Error qw(Error);
use DE_EPAGES::Object::API::Language qw(GetPKeyLanguageByCode);
use DE_EPAGES::Product::API::Constants qw(PRODUCT_ATTRIBUTE_TYPES); ##no critic (RequireCorrectImports)"
use DE_EPAGES::ShopMigration::API::Constants qw(
    PRODUCTTYPES_ATTRIBUTES_FOR_EXPORT
    PRODUCTTYPES_LOCALIZEDATTRIBUTES_FOR_EXPORT
    PRODUCTTYPESATTRIBUTE_ATTRIBUTES_FOR_EXPORT
    PRODUCTTYPESATTRIBUTE_LOCALIZEDATTRIBUTES_FOR_EXPORT
    PRODUCTTYPESATTRIBUTE_BOOLEAN_ATTRIBUTES_FOR_RESET
    PRODUCTTYPESATTRIBUTE_LOCALIZEDATTRIBUTES_FOR_RESET
    PREDEFATTRIBUTES_ATTRIBUTES_FOR_EXPORT
    PREDEFATTRIBUTES_LOCALIZEDATTRIBUTES_FOR_EXPORT
    PREDEFATTRIBUTES_OBJECT_ATTRIBUTES_FOR_EXPORT
    PRODUCTTYPE_SUPPORTED_ATTRIBUTE_TYPES
);
use DE_EPAGES::ShopMigration::API::ShopCommonData qw(GetActiveShopLanguages);
use DE_EPAGES::Object::API::Factory qw(LoadClassByAlias Transaction);
our @EXPORT_OK = qw(
    GetProductTypes
    SetProductTypes
    GetProductTypeData
    SetProductTypeData
);

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'PARENT_CATEGORY_NOTFOUND'      => 'Cannot create ProductType #Alias as parent #ParentPath was not found.',
        'UI_STORE_NOT_DEFINED'          => 'InstanceUnityUI is not defined in Database.d/#Store#SPACE[0].conf.',
        'INVALID_ELEMENT_FOR_ATTRIBUTE' => 'Invalid Element "#Name" found in #Type (Parent: #Parent)',
    }
}
#ERRORDEFINITION END

#========================================================================================
# function     GetProductTypes
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $ahProductTypes = GetProductTypes($Shop);
#----------------------------------------------------------------------------------------
# description  Fetches all relevant data for producttypes from the shop
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
#----------------------------------------------------------------------------------------
# return       $ahProductTypes | see GetProductTypesData() | reference to array of hashes
#========================================================================================
sub GetProductTypes {
    my ($Shop) = @_;

    my @Items;
    my $ProductTypes = $Shop->child('ProductTypes')->children();
    foreach my $ProductType ($Shop->get('ProductClass'), @$ProductTypes) {
        next unless $ProductType->class->alias eq 'ProductType';
        my $hInfo = GetProductTypeData($Shop, $ProductType);
        push @Items, $hInfo if (keys %$hInfo);
    }
    return \@Items;
}

#========================================================================================
# function     GetProductTypeData
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hProductType = GetProductTypeData($Shop, $ProductType);
#----------------------------------------------------------------------------------------
# description  Constructs a hash with the exportable fields for a single ProductType
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# input        $ProductType | product type | object
#----------------------------------------------------------------------------------------
# return       $hProductTypes | hash with following fields
#               <ul>
#                   <li>Attributes
#                       <ul>
#                           <li>Alias         - string (empty for ProductClass)</li>
#                           <li>Position      - float</li>
#                       </ul>
#                   </li>
#                   <li>LocalizedAttributes
#                       <ul>
#                           <li>LanguageCode2 - en|de|..
#                               <ul>
#                                   <li>Name - string</li>
#                               </ul>
#                           </li>
#                       </ul>
#                   </li>
#                   <li>ProductTypeAttributes
#                       <ul>
#                           <li>Attributes
#                               <ul>
#                                   <li>Alias               - string</li>
#                                   <li>IsVisible           - Boolean </li>
#                                   <li>IsHTML              - Boolean (reset only)</li>
#                                   <li>Position            - Float</li>
#                                   <li>Type                - String </li>
#                               </ul>
#                           </li>
#                           <li>LocalizedAttributes
#                               <ul>
#                                   <li>LanguageCode2       - en|de|..
#                                       <ul>
#                                           <li>Description - string (reset only)</li>
#                                           <li>Name        - string</li>
#                                       </ul>
#                                   </li>
#                               </ul>
#                           </li>
#                           <li>PreDefAttributes
#                               <ul>
#                                   <li>Alias - string
#                                       <ul>
#                                           <li>Attributes
#                                               <ul>
#                                                   <li>Alias              - string</li>
#                                                   <li>Position           - float</li>
#                                                   <li>ValueString        - string </li>
#                                                   <li>Attribute          - string</li>
#                                               </ul>
#                                           <li>LocalizedAttributes
#                                               <li>LanguageCode2          - en|de|..
#                                                   <ul>
#                                                       <li>ValueLocString - string</li>
#                                                   </ul>
#                                               </li>
#                                           </li>
#                                       </ul>
#                                   </li>
#                               </ul>
#                           </li>
#                       </ul>
#                   </li>
#               </ul>| reference to hash
#========================================================================================
sub GetProductTypeData {
    my ($Shop, $ProductType) = @_;

    my $Attributes = [ @{PRODUCTTYPES_ATTRIBUTES_FOR_EXPORT()} ];
    my $LocalizedAttributes = [ @{PRODUCTTYPES_LOCALIZEDATTRIBUTES_FOR_EXPORT()} ];
    my $hValues = { Attributes => $ProductType->get($Attributes), LocalizedAttributes => {}, };
    my $ProductClass = $Shop->get('ProductClass');
    $hValues->{Attributes}{Alias} = '' if ($ProductType->id == $ProductClass->id);
    my $ActiveLanguages = GetActiveShopLanguages($Shop);
    foreach my $hLanguage (@$ActiveLanguages) {
        my $AttrInfo = $ProductType->get($LocalizedAttributes, $hLanguage->{LanguageID});
        $hValues->{LocalizedAttributes}{$hLanguage->{Code2}} = $AttrInfo;
    }
    _convertScalarValuesToStringInHash($hValues);
    $hValues->{ProductTypeAttributes} = {};
    $Attributes = [ @{PRODUCTTYPESATTRIBUTE_ATTRIBUTES_FOR_EXPORT()} ];
    $LocalizedAttributes = [ @{PRODUCTTYPESATTRIBUTE_LOCALIZEDATTRIBUTES_FOR_EXPORT()} ];
    my @SupportedDefaultTypes = @{PRODUCTTYPE_SUPPORTED_ATTRIBUTE_TYPES()};
    $hValues->{ProductTypeAttributes} = [];
    foreach my $Attribute (sort { $a->alias cmp $b->alias } @{$ProductType->folder('Attributes')->children}) {
        my $AttributeInfo = {
            Attributes => $Attribute->get($Attributes), LocalizedAttributes => {}, PreDefAttributes => []
        };
        next if (!ElementOf($AttributeInfo->{Attributes}{Type}, @SupportedDefaultTypes));
        #$hValues->{ProductTypeAttributes}{$Alias} = $AttributeInfo;
        # Convert all attributes to LocalizedString (see UNITY-6516)
        my $Type = $AttributeInfo->{Attributes}{Type};
        $AttributeInfo->{Attributes}{Type} = $Type =~ /PreDef/ ? 'PreDefLocalizedString' : 'LocalizedString';
        foreach my $FieldName (@{PRODUCTTYPESATTRIBUTE_BOOLEAN_ATTRIBUTES_FOR_RESET()}) {
            $AttributeInfo->{Attributes}{$FieldName} = '0';
        }
        foreach my $hLanguage (@$ActiveLanguages) {
            my $AttrInfo = $Attribute->get($LocalizedAttributes, $hLanguage->{LanguageID});
            $AttributeInfo->{LocalizedAttributes}{$hLanguage->{Code2}} = $AttrInfo;
            foreach my $FieldName (@{PRODUCTTYPESATTRIBUTE_LOCALIZEDATTRIBUTES_FOR_RESET()}) {
                $AttributeInfo->{LocalizedAttributes}{$hLanguage->{'Code2'}}{$FieldName} = undef;
            }
        }
        _convertScalarValuesToStringInHash($AttributeInfo);
        if ($Attribute->existsChild('PreDefAttributes') && scalar @{$Attribute->child('PreDefAttributes')->children}) {
            my $PreDefAttributeInfo = {};
            my $PreDefAttributes = [ @{PREDEFATTRIBUTES_ATTRIBUTES_FOR_EXPORT()} ];
            my $PreDefLocalizedAttributes = [ @{PREDEFATTRIBUTES_LOCALIZEDATTRIBUTES_FOR_EXPORT()} ];
            foreach my $PredefAttribute (sort { $a->alias cmp $b->alias } @{$Attribute->child('PreDefAttributes')->children}) {
                $PreDefAttributeInfo = { Attributes => $PredefAttribute->get($PreDefAttributes), LocalizedAttributes => {}, };
                foreach my $ObjectAttribute (@{PREDEFATTRIBUTES_OBJECT_ATTRIBUTES_FOR_EXPORT()}) {
                    $PreDefAttributeInfo->{Attributes}{$ObjectAttribute} = $PredefAttribute->get($ObjectAttribute)->alias;
                }
                my $ValueString = $PredefAttribute->get('ValueString');
                foreach my $hLanguage (@$ActiveLanguages) {
                    my $AttrInfo = $PredefAttribute->get($PreDefLocalizedAttributes, $hLanguage->{LanguageID});
                    $PreDefAttributeInfo->{LocalizedAttributes}{$hLanguage->{Code2}} = $AttrInfo;
                    # Fallback if origin type is PreDefString which has only ValueString set, not ValueLocString
                    $PreDefAttributeInfo->{LocalizedAttributes}{$hLanguage->{Code2}}{ValueLocString} //= $ValueString;
                }
                _convertScalarValuesToStringInHash($PreDefAttributeInfo);
                push @{$AttributeInfo->{PreDefAttributes}}, $PreDefAttributeInfo;
            }
        }
        push @{$hValues->{ProductTypeAttributes}}, $AttributeInfo;
    }
    return $hValues;
}

#========================================================================================
# function     _convertScalarValuesToStringInHash
# state        private
#----------------------------------------------------------------------------------------
# syntax       _convertScalarValuesToStringInHash($hValues);
#----------------------------------------------------------------------------------------
# description  converts scalar values in the both hash lookups (Attributes and
#               LocalizedAttributes) to string values to have a consistent JSON quoting
#----------------------------------------------------------------------------------------
# input        $hValues         | Hash | ref hash
#========================================================================================
sub _convertScalarValuesToStringInHash {
    my ($hValues) = @_;

    # sometimes fields get treated as Integer instead of String thus mixing up the json file
    foreach my $Lookup (qw(Attributes LocalizedAttributes)) {
        foreach my $AttributeName (keys %{$hValues->{$Lookup}}) {
            if (defined $hValues->{$Lookup}{$AttributeName} and not ref $hValues->{$Lookup}{$AttributeName}) {
                $hValues->{$Lookup}{$AttributeName} .= ''
            }
        }
    }
}

#========================================================================================
# function     SetProductTypes
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetProductTypes($Shop, $ahProductTypes);
#----------------------------------------------------------------------------------------
# description  sets attributes on ProductTypes as specified in the $ahProductTypes
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# input        $hProductType    | see GetProductTypes() | reference to array of hashes
#========================================================================================
sub SetProductTypes {
    my ($Shop, $ahProductTypes) = @_;

    foreach my $hProductType (@$ahProductTypes) {
        SetProductTypeData($Shop, $hProductType);
    }
}

#========================================================================================
# function     SetProductTypeData
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetProductTypeData($Shop, $hProductType);
#----------------------------------------------------------------------------------------
# description  sets attributes on a ProductType as specified in the hash $hProductType
#----------------------------------------------------------------------------------------
# input        $Shop          | baseshop or prepshop | object
# input        $hProductType  | see GetProductTypes() | reference to array of hashes
#========================================================================================
sub SetProductTypeData {
    my ($Shop, $hProductType) = @_;

    my $Alias = $hProductType->{Attributes}{Alias};
    my $ProductType;
    my $ProductTypeFolder = $Shop->get('ProductTypes');
    my $BaseClass = $Shop->get('ProductClass');
    if ($Alias eq '') {
        $ProductType = $BaseClass;
        $Alias = 'ProductClass';
        delete $hProductType->{Attributes}{Alias};
    } else {
        $ProductType = $ProductTypeFolder->existsChild($Alias) ?
            $ProductTypeFolder->child($Alias) :
            $ProductTypeFolder->insertChild('ProductType', {
                Alias => $Alias, BaseClass => $BaseClass, Package => $BaseClass->get('Package'),
            });
    }
    my $Attributes = [ @{PRODUCTTYPES_ATTRIBUTES_FOR_EXPORT()} ];
    my $LocalizedAttributes = [ @{PRODUCTTYPES_LOCALIZEDATTRIBUTES_FOR_EXPORT()} ];
    my $TypeAttributes = [ @{PRODUCTTYPESATTRIBUTE_ATTRIBUTES_FOR_EXPORT()} ];
    my $TypeLocalizedAttributes = [ @{PRODUCTTYPESATTRIBUTE_LOCALIZEDATTRIBUTES_FOR_EXPORT()} ];
    my $PreDefAttributes = [ @{PREDEFATTRIBUTES_ATTRIBUTES_FOR_EXPORT()} ];
    my $PreDefLocalizedAttributes = [ @{PREDEFATTRIBUTES_LOCALIZEDATTRIBUTES_FOR_EXPORT()} ];
    my $ProductAttributeTypes = PRODUCT_ATTRIBUTE_TYPES();
    my $ProductAttributeClass = LoadClassByAlias('ProductAttribute');
    foreach my $Type (keys %$hProductType) {
        given ($Type) {
            when ('Attributes') {
                my $hParams = { Parent => "$Alias" };
                _checkAttributes($hProductType->{Attributes}, $Attributes, $hParams);
            }
            when ('LocalizedAttributes') {
                my $hParams = { IsLocalized => 1, Parent => "$Alias" };
                _checkAttributes($hProductType->{LocalizedAttributes}, $LocalizedAttributes, $hParams);
            }
            when ('ProductTypeAttributes') {
                foreach my $ProductTypeAttribute (@{$hProductType->{ProductTypeAttributes}}) {
                    my $ProductTypeAttributeAlias = $ProductTypeAttribute->{Attributes}{Alias};
                    my $hParams = { Parent => "$Alias/Attributes/$ProductTypeAttributeAlias" };
                    _checkAttributes($ProductTypeAttribute->{Attributes}, $TypeAttributes, $hParams);
                    $hParams->{IsLocalized} = 1;
                    _checkAttributes($ProductTypeAttribute->{LocalizedAttributes}, $TypeLocalizedAttributes, $hParams);
                    foreach my $PreDefAttribute (@{$ProductTypeAttribute->{PreDefAttributes}}) {
                        my $PreDefAttributeAlias = $PreDefAttribute->{Attributes}{Alias};
                        $hParams->{IsLocalized} = 0;
                        $hParams->{Parent} = "$Alias/Attributes/$ProductTypeAttributeAlias/PreDefAttributes/$PreDefAttributeAlias";
                        _checkAttributes($PreDefAttribute->{Attributes}, $PreDefAttributes, $hParams);
                        $hParams->{IsLocalized} = 1;
                        _checkAttributes($PreDefAttribute->{LocalizedAttributes}, $PreDefLocalizedAttributes, $hParams);
                    }
                }
            } default {
                die "Invalid Type: $Type";
            }
        }
    }

    Transaction(sub {
        foreach my $Type (keys %$hProductType) {
            given ($Type) {
                when ('Attributes') {
                    $ProductType->set($hProductType->{Attributes});
                }
                when ('LocalizedAttributes') {
                    foreach my $LanguageCode2 (keys %{$hProductType->{LocalizedAttributes}}) {
                        my $LanguageID = GetPKeyLanguageByCode($LanguageCode2);
                        if ($Shop->existsLanguage($LanguageID)) {
                            $ProductType->set($hProductType->{LocalizedAttributes}{$LanguageCode2}, $LanguageID);
                        }
                    }
                }
                when ('ProductTypeAttributes') {
                    foreach my $CurrentAttribute (@{$hProductType->{ProductTypeAttributes}}) {
                        my $ProductTypeAttributeAlias = $CurrentAttribute->{Attributes}{Alias};
                        my $AttributeFolder = $ProductType->folder('Attributes');
                        my $Package = $CurrentAttribute->{Attributes}{Type} eq 'LocalizedString' ?
                            $ProductAttributeTypes->{'LocalizedString'}{'Package'} :
                            $ProductAttributeTypes->{'PreDefLocalizedString'}{'Package'};
                        my $ProductAttribute = $AttributeFolder->existsChild($ProductTypeAttributeAlias) ?
                            $AttributeFolder->child($ProductTypeAttributeAlias) :
                            $AttributeFolder->insertChild('ProductAttribute ', {
                                Alias   => $ProductTypeAttributeAlias,
                                Package => $Package,
                                Type    => $CurrentAttribute->{Attributes}{Type},
                            });
                        $ProductAttribute->set($CurrentAttribute->{Attributes});
                        foreach my $LanguageCode2 (keys %{$CurrentAttribute->{LocalizedAttributes}}) {
                            my $LanguageID = GetPKeyLanguageByCode($LanguageCode2);
                            if ($Shop->existsLanguage($LanguageID)) {
                                $ProductAttribute->set($CurrentAttribute->{LocalizedAttributes}{$LanguageCode2}, $LanguageID);
                            }
                        }
                        next unless scalar @{$CurrentAttribute->{PreDefAttributes}};
                        my $PreDefFolder = $ProductAttribute->folder('PreDefAttributes');
                        foreach my $CurrentPreDef (@{$CurrentAttribute->{PreDefAttributes}}) {
                            my $PreDefAttributeAlias = $CurrentPreDef->{Attributes}{Alias};
                            my $PreDefAttribute = $PreDefFolder->existsChild($PreDefAttributeAlias) ?
                                $PreDefFolder->child($PreDefAttributeAlias) :
                                $PreDefFolder->insertChild('PreDefAttribute', { Alias => $PreDefAttributeAlias });
                            $CurrentPreDef->{Attributes}{PreDefClass} = $ProductAttributeClass;
                            $CurrentPreDef->{Attributes}{Attribute} = $ProductAttribute;
                            $PreDefAttribute->set($CurrentPreDef->{Attributes});
                            foreach my $LanguageCode2 (keys %{$CurrentPreDef->{LocalizedAttributes}}) {
                                my $LanguageID = GetPKeyLanguageByCode($LanguageCode2);
                                if ($Shop->existsLanguage($LanguageID)) {
                                    $PreDefAttribute->set($CurrentPreDef->{LocalizedAttributes}{$LanguageCode2}, $LanguageID);
                                }
                            }
                        }
                    }
                }
            }
        }
    });
}

#========================================================================================
# function     _checkKeysMatch
# state        private
#----------------------------------------------------------------------------------------
# syntax       _checkKeysMatch($Shop, $hProductType);
#----------------------------------------------------------------------------------------
# description  checks if all the keys in hash match with the corresponding
#               array
#----------------------------------------------------------------------------------------
# input        $hValues        | a hash with values | ref hash
# input        $aKeys          | list of allowed keys | ref array
# input        $hParams        | list of additional params | ref hash
#                The following parameters can be used:
#                       <ul>
#                           <li>IsLocalized - $hValues has sub hash for languages</li>
#                           <li>Type - a string with the type name of the hash</li>
#                           <li>Parent - a string with the parent alias of the hash</li>
#                       <ul>
#========================================================================================
sub _checkAttributes {
    my ($hValues, $aKeys, $hParams) = @_;

    $hParams->{IsLocalized} //= 0;
    $hParams->{Type} //= $hParams->{IsLocalized} ? 'LocalizedAttributes' : 'Attributes';
    $hParams->{Parent} //= 'ProductClass';
    if ($hParams->{IsLocalized}) {
        foreach my $LanguageCode2 (keys %$hValues) {
            foreach my $Key (keys %{$hValues->{$LanguageCode2}}) {
                if (!ElementOf($Key, @$aKeys)) {
                    Error('INVALID_ELEMENT_FOR_ATTRIBUTE', {
                        Name   => $Key,
                        Type   => $hParams->{Type},
                        Parent => $hParams->{Parent},
                    });
                }
            }
        }
    } else {
        foreach my $Key (keys %$hValues) {
            if (!ElementOf($Key, @$aKeys)) {
                Error('INVALID_ELEMENT_FOR_ATTRIBUTE', {
                    Name   => $Key,
                    Type   => $hParams->{Type},
                    Parent => $hParams->{Parent},
                });
            }
        }
    }
}

1;
