#========================================================================================
# package      DE_EPAGES::Spreedly::API::PaymentMethodGenerator
# state        public
# description  Class to generate new payment methods
#========================================================================================
package DE_EPAGES::Spreedly::API::PaymentMethodGenerator;

use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Object::API::Factory qw ( Transaction LoadObjectByPath LoadObjectByGUID ExistsObjectByPath);
use DE_EPAGES::Payment::API::Table::ShippingPaymentMap qw ( InsertShippingPaymentMap );
use DE_EPAGES::Core::API::PerlTools qw ( isValid GenerateShortGUID );
use DE_EPAGES::Spreedly::API::Constants qw ( IS_PAYMENT_METHOD_OK IS_PAYMENT_METHOD_NOK );

use strict;


#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( $Input1 );
#----------------------------------------------------------------------------------------
# description  constuctor
#----------------------------------------------------------------------------------------
# return       self | DE_EPAGES::Spreedly::API::PaymentMethodGenerator
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {
        'GeneratedPaymentMethods' => [],
    }, $class;
    return $self;
}

#========================================================================================
# function     generatePaymentMethods
# state        public
#----------------------------------------------------------------------------------------
# syntax       generatePaymentMethods( $PaymentMethod, $aSelectedPaymentMethods );
#----------------------------------------------------------------------------------------
# description  Creates new spreedly payment methods.
#----------------------------------------------------------------------------------------
# input        $PaymentMethod | spreedly payment method | PaymentMethodSpreedly
# input        $aSelectedPaymentMethods | selected payment methods to create | refarray
# return       $Status if creation was successful | boolean
#========================================================================================
sub generatePaymentMethods {
    my $self = shift;
    my ($PaymentMethod, $aSelectedPaymentMethods) = @_;

    return 0 unless isValid($PaymentMethod);
    return 0 unless isValid($aSelectedPaymentMethods);
    return 0 if (scalar @{ $aSelectedPaymentMethods }) < 1;

    my $GeneratePaymentMethodStatus;
    foreach my $PaymentTypeMethodGUID (@$aSelectedPaymentMethods) {
        $GeneratePaymentMethodStatus = $self->_generatePaymentMethod($PaymentMethod, $PaymentTypeMethodGUID);
        if($GeneratePaymentMethodStatus != IS_PAYMENT_METHOD_OK) {
            return IS_PAYMENT_METHOD_NOK;
        }
    }

    return IS_PAYMENT_METHOD_OK;
}

#========================================================================================
# function     _generatePaymentMethod
# state        private
#----------------------------------------------------------------------------------------
# syntax       _generatePaymentMethod( $PaymentTypeSpreedly, $PaymentMethod, $PaymentTypeMethodGUID );
#----------------------------------------------------------------------------------------
# description  Creates a new payment method by copying the given payment method
#               and changing the PaymentTypeMethod to the selected method.
#----------------------------------------------------------------------------------------
# input        $PaymentMethod | payment method to copy | PaymentMethodSpreedly
# input        $PaymentTypeMethodGUID | GUID of the select PaymentTypeMethod | String
# return       $Status | boolean
#========================================================================================
sub _generatePaymentMethod {
    my $self = shift;
    my ($PaymentMethod, $PaymentTypeMethodGUID) = @_;

    my $PaymentTypeMethod = $self->_LoadObjectByGUID($PaymentTypeMethodGUID);
    my $Shop = $PaymentMethod->getSite();

    my $NewPaymentMethodName = $PaymentTypeMethod->get('NameOrAlias');
    my $NewPaymentMethodAlias = $NewPaymentMethodName;

    my $hAliasCorrectionInfo = undef;
    my $AliasPath = '/Shops/' . $Shop->alias() . '/PaymentMethods/' . $NewPaymentMethodAlias;
    if ($self->_ExistsObjectByPath($AliasPath)) {
        my $OriginalNewPaymentMethodAlias = $NewPaymentMethodAlias;
        $NewPaymentMethodAlias = $NewPaymentMethodAlias . ' ' . GenerateShortGUID();
        $hAliasCorrectionInfo = {
            'OriginalAlias'     => $OriginalNewPaymentMethodAlias,
            'CorrectedAlias'    => $NewPaymentMethodAlias,
        };
    }

    # Parameters to copy from current payment type to new payment types
    my $hPaymentTypeParams = {
        'Alias' => $NewPaymentMethodAlias,
        'Name' => $NewPaymentMethodName,
        'Shop' => $Shop,
        'PaymentTypeMethod' => $PaymentTypeMethod,
        'Position' => 9999,
    };

    $hPaymentTypeParams = $self->_addCustomAttributes($PaymentMethod, $hPaymentTypeParams);

    eval {
        $self->_createNewPaymentMethod($PaymentMethod, $hPaymentTypeParams, $hAliasCorrectionInfo);
    };

    if (ExistsError()) {
        my $Error = GetError();
        warn ($Error->longMessage());
        return IS_PAYMENT_METHOD_NOK;
    }
    return IS_PAYMENT_METHOD_OK;
}

#========================================================================================
# function     _addCustomAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       _addCustomAttributes( $PaymentMethod, $hPaymentTypeParams );
#----------------------------------------------------------------------------------------
# description  Add custom attributes specific to each payment methods
#----------------------------------------------------------------------------------------
# input        $PaymentMethod | payment method | object
# input        $hPaymentTypeParams | Parameters of the payment type | ref_hash
# return       $hPaymentTypeParams | Parameters of the payment type | ref_hash
#========================================================================================
sub _addCustomAttributes {
    my $self = shift;
    my ($PaymentMethod, $hPaymentTypeParams) = @_;

    # Stripe Payment Intents attributes
    if($PaymentMethod->get('PaymentType')->alias() eq 'SpreedlyStripePaymentIntents') {
        my $hCredentials = $PaymentMethod->get([qw( StripeAccessToken StripeAccountId StripeRefreshToken )]);
        my %NewPaymentTypeParams = (%$hPaymentTypeParams, %$hCredentials);
        # return reference to merged hash
        return \%NewPaymentTypeParams;
    }

    return $hPaymentTypeParams;
}

#========================================================================================
# function     _createNewPaymentMethod
# state        private
#----------------------------------------------------------------------------------------
# syntax       _createNewPaymentMethod( $PaymentMethod, $hPaymentTypeParams, $hAliasCorrectionInfo );
#----------------------------------------------------------------------------------------
# description  Description
#----------------------------------------------------------------------------------------
# input        $PaymentMethod | payment method | object
# input        $hPaymentTypeParams | Parameters of the payment type | ref_hash
# input        $hAliasCorrectionInfo | Data of the Alias to be corrected | ref_hash
# return     
#========================================================================================
sub _createNewPaymentMethod {
    my $self = shift;
    my ($PaymentMethod, $hPaymentTypeParams, $hAliasCorrectionInfo) = @_;

    $self->_Transaction( sub {

        # create new payment method
        my $Shop = $hPaymentTypeParams->{'Shop'};
        my $PaymentTypeSpreedly = $self->_LoadObjectByPath('/PaymentTypes/' . $PaymentMethod->get('PaymentType')->alias());
        my $NewPaymentMethod = $PaymentTypeSpreedly->newPaymentMethod($Shop, $hPaymentTypeParams);

        my $hNewPaymentMethodInfo = {
            'NewPaymentMethod'      => $NewPaymentMethod,
            'NewPaymentMethodID'    => $NewPaymentMethod->id(),
            'PaymentTypeMethodName' => $NewPaymentMethod->get('PaymentTypeMethod')->get('NameOrAlias'),
            'AliasCorrection'       => $hAliasCorrectionInfo,
            'NewPaymentMethodName'  => $NewPaymentMethod->alias()
        };
        push(@{$self->{GeneratedPaymentMethods}}, $hNewPaymentMethodInfo);

        # activate on all shipping methods (see MBO > delivery > dependencies)
        foreach my $ShippingMethod (@{$Shop->get('ShippingMethods')}) {
            $self->_InsertShippingPaymentMap({
                'ShippingMethodID'  => $ShippingMethod->id(),
                'PaymentMethodID'   => $NewPaymentMethod->id(),
            });
        }
        $NewPaymentMethod->setInShopNameLikePaymentTypeMethod();
        $NewPaymentMethod->uploadPaymentTypeMethodLogo();
    });
    return;
}


#========================================================================================
# function     getGeneratedPaymentMethods
# state        public
#----------------------------------------------------------------------------------------
# syntax       getGeneratedPaymentMethods( );
#----------------------------------------------------------------------------------------
# description  Returns the generated payment methods
#----------------------------------------------------------------------------------------
# return       the generated payment methods | refarray
#========================================================================================
sub getGeneratedPaymentMethods {
    my $self = shift;
    return $self->{'GeneratedPaymentMethods'};
}

#========================================================================================
# function     _LoadObjectByPath
# state        private
#----------------------------------------------------------------------------------------
# syntax       _LoadObjectByPath( $Path, $RootObjectID );
#----------------------------------------------------------------------------------------
# description  Wrapper of core binding
#----------------------------------------------------------------------------------------
# input        $Path | path of alias names from the root object | string
# return       $Object | object | object
#========================================================================================
sub _LoadObjectByPath {
    my $self = shift;
    my ($Path) = @_;
    return LoadObjectByPath($Path);
}

#========================================================================================
# function     _LoadObjectByGUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       _LoadObjectByGUID( $GUID );
#----------------------------------------------------------------------------------------
# description  Wrapper of core binding
#----------------------------------------------------------------------------------------
# input        $GUID | general unique id of the object | string
# return       $Object | object  | object
#========================================================================================
sub _LoadObjectByGUID {
    my $self = shift;
    my ($GUID) = @_;
    return LoadObjectByGUID($GUID);
}

#========================================================================================
# function     _ExistsObjectByPath
# state        private
#----------------------------------------------------------------------------------------
# syntax       _ExistsObjectByPath( $Path, $RootObjectID );
#----------------------------------------------------------------------------------------
# description  Wrapper of core binding
#----------------------------------------------------------------------------------------
# input        $Path | path of alias names from the root object | String
# input        $RootObjectID | objectId from the root object | int
# return       true if the specified object exists | boolean
#========================================================================================
sub _ExistsObjectByPath {
    my $self = shift;
    my ($Path, $RootObjectID) = @_;
    return ExistsObjectByPath($Path, $RootObjectID);
}

#========================================================================================
# function     _Transaction
# state        private
#----------------------------------------------------------------------------------------
# syntax       _Transaction( $cCode, $cRollback );
#----------------------------------------------------------------------------------------
# description  Wrapper of core binding
#----------------------------------------------------------------------------------------
# input        $cCode | code block | code reference
# input        $cRollback | roll back code block (optional) | code reference
#========================================================================================
sub _Transaction {
    my $self = shift;
    my ($cCode, $cRollback) = @_;
    Transaction($cCode, $cRollback);
    return;
}

#========================================================================================
# function     _InsertShippingPaymentMap
# state        private
#----------------------------------------------------------------------------------------
# syntax       _InsertShippingPaymentMap($hShippingPaymentMap);
#----------------------------------------------------------------------------------------
# description  Wrapper of core binding
#----------------------------------------------------------------------------------------
# input        $hShippingPaymentMap
#========================================================================================
sub _InsertShippingPaymentMap {
    my $self = shift;
    my ($hShippingPaymentMap) = @_;
    InsertShippingPaymentMap($hShippingPaymentMap);
    return;
}

1;