use strict;
use warnings;
use Getopt::Long;
use Time::HiRes qw( time );

use DE_EPAGES::Cartridge::API::Table::Cartridge qw ( ExistsCartridgeByPackage );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Core::API::Error qw( GetError ExistsError );
use DE_EPAGES::Core::API::Log qw( LogError );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Factory qw( LoadObject LoadClassByAlias LoadObjectByPath);
use DE_EPAGES::Spreedly::API::Cartridge;
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );

use constant PAYMENT_METHOD_STRIPE_PI => 'PaymentMethodSpreedlyStripePaymentIntents';

sub Main {
    local $| = 1;

    my $Password;
    my $StoreName;
    my $ShopAlias;
    my $Help;
    my $NoOutput;

    GetOptions(
        'help'        => \$Help,
        'nooutput'    => \$NoOutput,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'shop=s'      => \$ShopAlias,
    );

    usage() if $Help;

    unless ( defined $StoreName ) {
        print STDERR "Missing parameter -storename.\n\n";
        usage();
    }

    my $time1 = time;
    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub {
            if (not ExistsCartridgeByPackage('DE_EPAGES::Spreedly')){
                print 'Store: '.$StoreName.": Cartridge Spreedly is not installed.\n" unless ($NoOutput);
                return;
            }

            if (isValid($ShopAlias)) {
                my $Shop;
                eval { $Shop = LoadObjectByPath('/Shops')->child($ShopAlias); };
                if (ExistsError()) {
                    printf STDERR "[%s] WARN: Shop '%s' does not exist.\n", $StoreName, $ShopAlias;
                    my $Error = GetError();
                    print("failed with error \"" . $Error->longMessage() . "\".\n\n");
                } else {
                    # fix payment method attribute only for this shop
                    fixPaymentAttributes($NoOutput, $StoreName, $Shop);
                }
            } else {
                # fix for all stripe pi payment methods from store
                fixPaymentAttributes($NoOutput, $StoreName);
            }

        },
    );
    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

#========================================================================================
# function         fixPaymentAttributes
#----------------------------------------------------------------------------------------
# syntax           fixPaymentAttributes($NoOutput, $StoreName, $Shop);
#----------------------------------------------------------------------------------------
# description      Get shop payment methods
#                   Validate if the fix should be applied
#                   Copy StripeAccountId from the first valid payment method to the other
#                   payments StripePaymentIntents that are missing this attribute
#========================================================================================
sub fixPaymentAttributes {
    my ($NoOutput, $StoreName, $Shop) = @_;
    my $hShopsWithPaymentMethods;

    if(defined $Shop) {
        my $ShopAlias = $Shop->alias();
        print 'Running on store: '.$StoreName." only for shop: ".$ShopAlias."\n" unless ($NoOutput);
        $hShopsWithPaymentMethods = _getStripePaymentMethodsByShop($NoOutput, $Shop);
    } else {
        print 'Running on store: '.$StoreName." for all shops.\n" unless ($NoOutput);
        $hShopsWithPaymentMethods = _getStripePaymentMethodsFromStore($NoOutput, $StoreName);
    }

    foreach my $ShopAlias (keys %{ $hShopsWithPaymentMethods }) {

        print "------------------------\n"  unless ($NoOutput);
        print $ShopAlias."\n"  unless ($NoOutput);

        my $hInstalledPaymentMethods = $hShopsWithPaymentMethods->{$ShopAlias};

        eval {
            my $ValidStripePaymentMethod;
            if(_validateInstalledPayments($NoOutput, $hInstalledPaymentMethods, $StoreName, $ShopAlias))
            {
                print "\tCopy StripeAccountId attribute to the payment methods with empty attribute \n" unless ($NoOutput);
                my $MethodsWithAttributeSet = $hInstalledPaymentMethods->{'WithAttributeSet'};
                my $MethodsWithAttributeEmpty = $hInstalledPaymentMethods->{'WithAttributeEmpty'};
                print "\t SpreedlyStripePaymentIntents payment method with StripeAccountId attribute set: \n" unless ($NoOutput);
                foreach my $PaymentMethodId (keys %{ $MethodsWithAttributeSet }) {
                    print "\t - ID: ".$PaymentMethodId." - StripeAccountId: ".$MethodsWithAttributeSet->{$PaymentMethodId}->{'StripeAccountId'}
                        ." PaymentMethod:".$MethodsWithAttributeSet->{$PaymentMethodId}->{'PaymentMethodAlias'}."\n"  unless ($NoOutput);
                    $ValidStripePaymentMethod =  LoadObject($PaymentMethodId);
                    last;
                }

                if(defined $ValidStripePaymentMethod) {
                    my $hStripeCredentials = $ValidStripePaymentMethod->get([ qw ( StripeAccountId StripeAccessToken StripeRefreshToken ) ]);
                    print "\t Stripe account id: ".$hStripeCredentials->{'StripeAccountId'}."\n" unless ($NoOutput);
                    foreach my $PaymentMethodId (keys %{ $MethodsWithAttributeEmpty }) {
                        my $PaymentMethodAttributeEmpty = LoadObject($PaymentMethodId);
                        $PaymentMethodAttributeEmpty->set(  {
                            'StripeAccountId' => $hStripeCredentials->{'StripeAccountId'},
                            'StripeAccessToken' => $hStripeCredentials->{'StripeAccessToken'},
                            'StripeRefreshToken' => $hStripeCredentials->{'StripeRefreshToken'}
                        });
                        print "\t Updated payment method Id: ".$PaymentMethodAttributeEmpty->id." Alias: ".$PaymentMethodAttributeEmpty->alias()."\n" unless ($NoOutput);

                    }
                }
            }
        };
        if (ExistsError()) {
            warn('Patch failed in shop ' . $ShopAlias . " with error \"" . GetError() . "\".")
        } else {
            print "\t Done. \n" unless ($NoOutput);
        }
    }
    print "\n" unless ($NoOutput);

    eval { SynchronizeCache(); };
    LogError(GetError()) if ExistsError();

}

#========================================================================================
# function         _getStripePaymentMethodsByShop
#----------------------------------------------------------------------------------------
# syntax           _getStripePaymentMethodsByShop($NoOutput, $Shop);
#----------------------------------------------------------------------------------------
# description      Read SpreedlyStripePaymentIntents payment methods from shop.
#                   Group them by shop, methods with correct attributes, methods with
#                   missing attributes. Keep a counter of all payments per shop.
#========================================================================================
sub _getStripePaymentMethodsByShop {
    my ($NoOutput, $Shop) = @_;
    my %InstalledPaymentMethods = ();

    if($Shop->featureMaxValue('SpreedlyStripePaymentIntents') == 0) {
        print "\t Feature SpreedlyStripePaymentIntents is NOT installed.\n" unless ($NoOutput);
        print "\t Shop skipped. \n" unless ($NoOutput);
        return;
    }

    my $ShopAlias = $Shop->alias();

    my $PaymentMethods = $Shop->get('PaymentMethods');
    foreach my $PaymentMethod (@$PaymentMethods){
        if ($PaymentMethod->class()->alias() eq 'PaymentMethodSpreedlyStripePaymentIntents') {
            $InstalledPaymentMethods{$ShopAlias}{'CountStripePIMethods'}++;
            my $StripeAccountId = $PaymentMethod->get('StripeAccountId');
            my $PaymentMethodId = $PaymentMethod->id;
            my $hPaymentValues = {
                ShopAlias => $ShopAlias,
                PaymentMethodAlias => $PaymentMethod->alias(),
                StripeAccountId    => undef
            };
            if(isValid($StripeAccountId)) {
                $hPaymentValues->{'StripeAccountId'} = $StripeAccountId;
                $InstalledPaymentMethods{$ShopAlias}{'WithAttributeSet'}{$PaymentMethodId} = $hPaymentValues;
            } else {
                $InstalledPaymentMethods{$ShopAlias}{'WithAttributeEmpty'}{$PaymentMethodId} = $hPaymentValues;
            }
        }
    }


    return \%InstalledPaymentMethods;
}

#========================================================================================
# function         _getStripePaymentMethodsFromStore
#----------------------------------------------------------------------------------------
# syntax           _getStripePaymentMethodsFromStore();
#----------------------------------------------------------------------------------------
# description      Read SpreedlyStripePaymentIntents payment methods from store.
#                   Group them by shop, methods with correct attributes, methods with
#                   missing attributes. Keep a counter of all payments per shop.
#========================================================================================
sub _getStripePaymentMethodsFromStore {
    my %InstalledPaymentMethods = ();

    my $SpreedlyClass = LoadClassByAlias(PAYMENT_METHOD_STRIPE_PI);
    my $aPaymentMethodIds = $SpreedlyClass->findObjects('ByClassID',[$SpreedlyClass->id]);
    foreach my $PaymentMethodId (@{$aPaymentMethodIds}){
        my $PaymentMethod = LoadObject($PaymentMethodId);
        my $ShopAlias = $PaymentMethod->get('Shop')->alias();
        my $StripeAccountId = $PaymentMethod->get('StripeAccountId');
        $InstalledPaymentMethods{$ShopAlias}{'CountStripePIMethods'}++;

        my $hPaymentValues = {
            ShopAlias => $ShopAlias,
            PaymentMethodAlias => $PaymentMethod->alias(),
            StripeAccountId    => undef
        };
        if(isValid($StripeAccountId)) {
            $hPaymentValues->{'StripeAccountId'} = $StripeAccountId;
            $InstalledPaymentMethods{$ShopAlias}{'WithAttributeSet'}{$PaymentMethodId} = $hPaymentValues;
        } else {
            $InstalledPaymentMethods{$ShopAlias}{'WithAttributeEmpty'}{$PaymentMethodId} = $hPaymentValues;
        }

    }

    return \%InstalledPaymentMethods;
}

#========================================================================================
# function         _validateInstalledPayments
#----------------------------------------------------------------------------------------
# syntax           _validateInstalledPayments($NoOutput, $hInstalledPaymentMethods);
#----------------------------------------------------------------------------------------
# description      Check different cases when payment methods exists, not exist, have
#                   correct attributes, empty attributes.
#                   Return 0 in case of error or when the patch can not be applied.
#                   Return 1 when the process can continue and the patch can be applied.
#========================================================================================
sub _validateInstalledPayments {
    my ($NoOutput, $hInstalledPaymentMethods, $StoreName, $ShopAlias) = @_;

    if($hInstalledPaymentMethods->{'CountStripePIMethods'} == 0) {
        printf("\t [%s] INFO: No payment methods of type SpreedlyStripePaymentIntents installed. Skipped Shop '%s' .\n", $StoreName, $ShopAlias) unless ($NoOutput);
        return 0;
    }

    my $MethodsWithAttributeSet = $hInstalledPaymentMethods->{'WithAttributeSet'};
    my $MethodsWithAttributeEmpty = $hInstalledPaymentMethods->{'WithAttributeEmpty'};

    if(scalar keys(%{ $MethodsWithAttributeEmpty })  == 0) {
        printf("\t [%s] INFO:  No payment methods with empty stripe payment intents attributes. Skipped Shop '%s' .\n", $StoreName, $ShopAlias) unless ($NoOutput);
        return 0;
    }

    if((scalar keys(%{ $MethodsWithAttributeSet })  == 0) && (scalar keys(%{ $MethodsWithAttributeEmpty }) > 0)){
        printf STDERR "[%s] WARN: Found only payment methods with invalid StripeAccountId. Either these payment methods were not onboarded, or method was duplicated and the initial method was deleted. Skipped Shop '%s' .\n", $StoreName, $ShopAlias;
        foreach my $KeyId (keys %{ $MethodsWithAttributeEmpty }) {
            print "\t - ".$KeyId." - ".$MethodsWithAttributeEmpty->{$KeyId}{'PaymentMethodAlias'}."\n" unless ($NoOutput);
        }
        return 0;
    }

    if(scalar keys(%{ $MethodsWithAttributeSet })  == 0) {
        printf STDERR "[%s] WARN: Payment method 'SpreedlyStripePaymentIntents' with valid StripeAccountId was NOT found. Skipped Shop '%s' .\n", $StoreName, $ShopAlias;
        return 0;
    }

    if(scalar keys(%{ $MethodsWithAttributeSet })  > 1) {
        my %AccountIds;
        foreach my $KeyId (keys %{ $MethodsWithAttributeSet }) {
            $AccountIds{$MethodsWithAttributeSet->{$KeyId}{'StripeAccountId'}} = $MethodsWithAttributeSet->{$KeyId}{'StripeAccountId'};
        }

        if(keys(%AccountIds)  > 1) {
            printf STDERR "[%s] WARN: Payment method 'SpreedlyStripePaymentIntents' is installed with two or more different StripeAccountIds. Skipped Shop '%s' .\n", $StoreName, $ShopAlias;
            foreach my $KeyId (keys %{ $MethodsWithAttributeSet }) {
                print "\t - ".$KeyId." - ".$MethodsWithAttributeSet->{$KeyId}{'PaymentMethodAlias'}." - ".$MethodsWithAttributeSet->{$KeyId}{'StripeAccountId'}."\n" unless ($NoOutput);
            }
            return 0;
        }
    }

    print "\t SpreedlyStripePaymentIntents payment methods with empty attributes: \n" unless ($NoOutput);
    unless ($NoOutput) {
        foreach my $KeyId (keys %{ $MethodsWithAttributeEmpty }) {
            print "\t - ".$KeyId." - Missing StripeAccountId: ".$MethodsWithAttributeEmpty->{$KeyId}{'PaymentMethodAlias'}."\n";
        }
    }
    print "\n" unless ($NoOutput);

    return 1;
}


sub usage {
    print <<END_USAGE;
Description:
    Get all payment methods Stripe Payment Intents from store, grouped by shop,
    copy StripeAccountId and Stripe tokens from the first valid method to all other
    payment methods of type SpreedlyStripePaymentIntents with empty attributes.
    Log messages with a WARN prefix are shown when the script cannot fix
    the payment method and the Merchant should be informed to do the on-boarding again.

Usage:
    perl $0 [options] [flags]
options:
    -passwd     database user password                          ()
    -storename  name of store                                   ()
    -shop       only run the script for this shop               ()
flags:
    -help       show the command line options
    -nooutput   do not print any output
Example:
    perl $0 -storename Store
    perl $0 -storename Store -shop DemoShop
    perl $0 -storename Store -shop DemoShop -nooutput
END_USAGE
    exit 2;
}

#========================================================================================
# SCRIPT RUN
#========================================================================================
RunScript('Sub' => \&Main);