#========================================================================================
# package      DE_EPAGES::Spreedly::t::unit::API_PaymentMethodGenerator
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::Spreedly::API::PaymentMethodGenerator
#========================================================================================
package DE_EPAGES::Spreedly::t::unit::API_PaymentMethodGenerator;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Test::Mock::EpagesObject;
use Test::Exception;
use Test::More;
use DE_EPAGES::Spreedly::API::PaymentMethodGenerator;
use Hash::Merge qw( merge );
use DE_EPAGES::Core::API::PerlTools qw ( GenerateShortGUID );

use constant IS_NAME_OR_ALIAS => 1;
use constant STRIPE_PI_TYPE  => 'SpreedlyStripePaymentIntents';
use strict;


#----------------------------------------------------------------------------------------
sub test {
    my $self = shift;
    $self->test_PaymentTypeGenerationWithUndefParams();
    $self->test_PaymentTypeGenerationWithEmptyPaymentMethodList();
    $self->test_PaymentTypeGeneration();
    $self->test_PaymentTypeGenerationStripe();
    $self->test_PaymentTypeGenerationWithAlreadyExistingAlias();
    $self->test_PaymentTypeGenerationWithError();
    done_testing();
}

#----------------------------------------------------------------------------------------
sub test_PaymentTypeGenerationWithUndefParams {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $PaymentMethodGenerator = DE_EPAGES::Spreedly::API::PaymentMethodGenerator->new();
    my $PaymentMethod = undef;
    my $aSelectedPaymentMethods = undef;

    my $Status = $PaymentMethodGenerator->generatePaymentMethods( $PaymentMethod, $aSelectedPaymentMethods );
    is($Status, 0,  "$SubTestName - Tests if status flag is 0 if inputs are undef");

    my $aGeneratedPaymentMethods = $PaymentMethodGenerator->getGeneratedPaymentMethods();
    is_deeply($aGeneratedPaymentMethods, [], "$SubTestName - Tests if return array us empty if inputs are undef");
}

#----------------------------------------------------------------------------------------
sub test_PaymentTypeGenerationWithEmptyPaymentMethodList {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $PaymentMethodGenerator = DE_EPAGES::Spreedly::API::PaymentMethodGenerator->new();
    my $PaymentMethod = $self->_getPaymentMethodSpreedly();
    my $aSelectedPaymentMethods = [];

    my $Status = $PaymentMethodGenerator->generatePaymentMethods( $PaymentMethod, $aSelectedPaymentMethods );
    is($Status, 0,  "$SubTestName - Tests if status flag is 0 if input array is empty");

    my $aGeneratedPaymentMethods = $PaymentMethodGenerator->getGeneratedPaymentMethods();
    is_deeply($aGeneratedPaymentMethods, [], "$SubTestName - Tests if return array us empty if input array is empty");
}

#----------------------------------------------------------------------------------------
sub test_PaymentTypeGeneration {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $PaymentMethodGenerator = DE_EPAGES::Spreedly::API::PaymentMethodGenerator->new();
    $PaymentMethodGenerator = Test::MockObject::Extends->new($PaymentMethodGenerator);

    $PaymentMethodGenerator->mock('_LoadObjectByPath', sub {
        return $self->_getPaymentTypeSpreedly();
    });
    $PaymentMethodGenerator->mock('_LoadObjectByGUID', sub {
        return $self->_getPaymentTypeMethod(IS_NAME_OR_ALIAS);
    });
    $PaymentMethodGenerator->mock('_ExistsObjectByPath', sub {
        return 0;
    });
    $PaymentMethodGenerator->mock('_Transaction', sub {
        my $self = shift;
        my ($FunctionPointer) = @_;
        $FunctionPointer->();
        return;
    });

    my $PaymentMethod = $self->_getPaymentMethodSpreedly();
    my $aSelectedPaymentMethods = ['314'];

    my $Status = $PaymentMethodGenerator->generatePaymentMethods( $PaymentMethod, $aSelectedPaymentMethods );
    is($Status, 1, "$SubTestName - Tests if status flag is 1 if valid inputs are given");

    my $aGeneratedPaymentMethods = $PaymentMethodGenerator->getGeneratedPaymentMethods();
    is(scalar @{ $aGeneratedPaymentMethods }, scalar @{ $aSelectedPaymentMethods }, "$SubTestName - Is number of generated payment method equal to number of selected");

    my $GeneratedPaymentMethod = $aGeneratedPaymentMethods->[0];
    is($GeneratedPaymentMethod->{'PaymentTypeMethodName'}, 'MyPaymentTypeMethod', "$SubTestName - test if payment type method of new type is correct");
    is($GeneratedPaymentMethod->{'NewPaymentMethod'}->get('Name'), 'MyPaymentTypeMethod', "$SubTestName - test if new payment has correct name");
    is($GeneratedPaymentMethod->{'NewPaymentMethod'}->alias(), 'MyPaymentTypeMethod', "$SubTestName - test if new payment has correct name");

}

#----------------------------------------------------------------------------------------
sub test_PaymentTypeGenerationWithAlreadyExistingAlias {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $PaymentMethodGenerator = DE_EPAGES::Spreedly::API::PaymentMethodGenerator->new();
    $PaymentMethodGenerator = Test::MockObject::Extends->new($PaymentMethodGenerator);

    $PaymentMethodGenerator->mock('_LoadObjectByPath', sub {
        return $self->_getPaymentTypeSpreedly();
    });
    $PaymentMethodGenerator->mock('_LoadObjectByGUID', sub {
        return $self->_getPaymentTypeMethod(IS_NAME_OR_ALIAS);
    });
    $PaymentMethodGenerator->mock('_ExistsObjectByPath', sub {
        return 1;
    });
    $PaymentMethodGenerator->mock('_Transaction', sub {
        my $self = shift;
        my ($FunctionPointer) = @_;
        $FunctionPointer->();
        return;
    });

    my $PaymentMethod = $self->_getPaymentMethodSpreedly();
    my $aSelectedPaymentMethods = ['314'];

    my $Status = $PaymentMethodGenerator->generatePaymentMethods( $PaymentMethod, $aSelectedPaymentMethods );
    is($Status, 1, "$SubTestName - Tests if status flag is 1 if valid inputs are given");

    my $aGeneratedPaymentMethods = $PaymentMethodGenerator->getGeneratedPaymentMethods();
    is(scalar @{ $aGeneratedPaymentMethods }, scalar @{ $aSelectedPaymentMethods }, "$SubTestName - Is number of generated payment method equal to number of selected");

    my $GeneratedPaymentMethod = $aGeneratedPaymentMethods->[0];
    is($GeneratedPaymentMethod->{'PaymentTypeMethodName'}, 'MyPaymentTypeMethod', "$SubTestName - test if payment type method of new type is correct");
    is($GeneratedPaymentMethod->{'NewPaymentMethod'}->get('Name'), 'MyPaymentTypeMethod', "$SubTestName - test if new payment has correct name");

    my $NewPaymentMethodAlias = $GeneratedPaymentMethod->{'NewPaymentMethod'}->alias();
    my $NewPaymentMethodAliasLength = length($NewPaymentMethodAlias);
    my $NewPaymentMethodNameWithGuid = $GeneratedPaymentMethod->{'NewPaymentMethod'}->get('Name').' '.GenerateShortGUID();
    my $NewPaymentMethodNameWithGuidLength = length($NewPaymentMethodNameWithGuid);

    is($NewPaymentMethodAliasLength, $NewPaymentMethodNameWithGuidLength, "$SubTestName - test if the alias is built from the name plus a GUID");

}

#----------------------------------------------------------------------------------------
sub test_PaymentTypeGenerationWithError {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $PaymentMethodGenerator = DE_EPAGES::Spreedly::API::PaymentMethodGenerator->new();
    $PaymentMethodGenerator = Test::MockObject::Extends->new($PaymentMethodGenerator);

    $PaymentMethodGenerator->mock('_LoadObjectByPath', sub {
        return $self->_getPaymentTypeSpreedly();
    });
    $PaymentMethodGenerator->mock('_LoadObjectByGUID', sub {
        return $self->_getPaymentTypeMethod(!IS_NAME_OR_ALIAS);
    });
    $PaymentMethodGenerator->mock('_ExistsObjectByPath', sub {
        return 0;
    });
    $PaymentMethodGenerator->mock('_Transaction', sub {
        die();
    });
    my $PaymentMethod = $self->_getPaymentMethodSpreedly();
    my $aSelectedPaymentMethods = ['314'];

    my $Status = $PaymentMethodGenerator->generatePaymentMethods( $PaymentMethod, $aSelectedPaymentMethods );
    is($Status, 0, "$SubTestName - Tests if status flag is 1 if valid inputs are given");
}

#----------------------------------------------------------------------------------------
sub test_PaymentTypeGenerationStripe {
    my $self = shift;
    my $SubTestName = (caller(0))[3];

    my $PaymentMethodGenerator = DE_EPAGES::Spreedly::API::PaymentMethodGenerator->new();
    $PaymentMethodGenerator = Test::MockObject::Extends->new($PaymentMethodGenerator);

    my $MockedPaymentTypeSpreedly =  $self->_getPaymentTypeSpreedly(STRIPE_PI_TYPE);
    $PaymentMethodGenerator->mock('_LoadObjectByPath', sub {
        return $MockedPaymentTypeSpreedly;
    });
    $PaymentMethodGenerator->mock('_LoadObjectByGUID', sub {
        return $self->_getPaymentTypeMethod(IS_NAME_OR_ALIAS);
    });
    $PaymentMethodGenerator->mock('_ExistsObjectByPath', sub {
        return 0;
    });
    $PaymentMethodGenerator->mock('_Transaction', sub {
        my $self = shift;
        my ($FunctionPointer) = @_;
        $FunctionPointer->();
        return;
    });

    my $hStripeAttributes = {
        'StripeAccessToken'  => 'SomeStripeAccessToken',
        'StripeAccountId'    => 'SomeStripeAccountId',
        'StripeRefreshToken' => 'SomeStripeRefreshToken'
    };

    my $PaymentMethod = $self->_getPaymentMethodSpreedly($hStripeAttributes, STRIPE_PI_TYPE);
    my $aSelectedPaymentMethods = ['314'];

    my $Status = $PaymentMethodGenerator->generatePaymentMethods( $PaymentMethod, $aSelectedPaymentMethods );
    is($Status, 1, "$SubTestName - Tests if status flag is 1 if valid inputs are given");

    is( $MockedPaymentTypeSpreedly->call_pos(1),  'newPaymentMethod', "$SubTestName - test if newPaymentMethod is called" );

    is( ($MockedPaymentTypeSpreedly->call_args(1))[2]->{'Position'}, '9999', "$SubTestName - test if passed argument Position is correct" );
    is( ($MockedPaymentTypeSpreedly->call_args(1))[2]->{'StripeAccessToken'}, 'SomeStripeAccessToken', "$SubTestName - test if passed argument StripeAccessToken is correct" );
    is( ($MockedPaymentTypeSpreedly->call_args(1))[2]->{'StripeAccountId'}, 'SomeStripeAccountId', "$SubTestName - test if passed argument StripeAccountId is correct" );
    is( ($MockedPaymentTypeSpreedly->call_args(1))[2]->{'StripeRefreshToken'}, 'SomeStripeRefreshToken', "$SubTestName - test if passed argument StripeRefreshToken is correct" );

    my $aGeneratedPaymentMethods = $PaymentMethodGenerator->getGeneratedPaymentMethods();
    is(scalar @{ $aGeneratedPaymentMethods }, scalar @{ $aSelectedPaymentMethods }, "$SubTestName - Is number of generated payment method equal to number of selected");

    my $GeneratedPaymentMethod = $aGeneratedPaymentMethods->[0];
    is($GeneratedPaymentMethod->{'PaymentTypeMethodName'}, 'MyPaymentTypeMethod', "$SubTestName - test if payment type method of new type is correct");
    is($GeneratedPaymentMethod->{'NewPaymentMethod'}->get('Name'), 'MyPaymentTypeMethod', "$SubTestName - test if new payment has correct name");
    is($GeneratedPaymentMethod->{'NewPaymentMethod'}->alias(), 'MyPaymentTypeMethod', "$SubTestName - test if new payment has correct name");

}

#----------------------------------------------------------------------------------------
sub _getPaymentTypeSpreedly {
    my $self = shift;
    my($PaymentTypeAlias) = @_;
    my $hAttributes = {
        'Name'  => 'Test Payment Type',
        'Alias' => $PaymentTypeAlias // 'Test Payment Type Alias'
    };
    my $MockedPaymentTypeSpreedly = DE_EPAGES::Test::Mock::EpagesObject->new(
        'DE_EPAGES::Payment::API::Object::PaymentType', $hAttributes
    );
    $MockedPaymentTypeSpreedly->mock('newPaymentMethod', sub {
        shift;
        my ($Shop, $hParameters) = @_;
        return $self->_getPaymentMethodSpreedly($hParameters);
    });
    return $MockedPaymentTypeSpreedly;
}

#----------------------------------------------------------------------------------------
sub _getPaymentMethodSpreedly {
    my $self = shift;
    my ($hParameters, $PaymentTypeAlias) = @_;

    my $hAttributes = {
        'ID'                => '123',
        'PaymentType'       => $self->_getPaymentTypeSpreedly($PaymentTypeAlias),
        'PaymentTypeMethod' => $self->_getPaymentTypeMethod(IS_NAME_OR_ALIAS)
    };

    $hAttributes = merge($hAttributes, $hParameters);

    my $MockedPaymentMethodSpreedly = DE_EPAGES::Test::Mock::EpagesObject->new(
        'DE_EPAGES::Spreedly::API::Object::PaymentMethodSpreedly', $hAttributes
    );
    $MockedPaymentMethodSpreedly->mock('getSite', sub {
        return $self->_getSite();
    });
   $MockedPaymentMethodSpreedly->mock('setInShopNameLikePaymentTypeMethod', sub {
        return;
    });
   $MockedPaymentMethodSpreedly->mock('uploadPaymentTypeMethodLogo', sub {
        return;
    });

    return $MockedPaymentMethodSpreedly;
}

#----------------------------------------------------------------------------------------
sub _getSite {
    my $self = shift;
    my $hAttributes = {
        'Alias'             => 'MyShop',
        'ShippingMethods'   => [],
    };
    my $MockedSite = DE_EPAGES::Test::Mock::EpagesObject->new(
        'DE_EPAGES::Shop::API::Object::Shop', $hAttributes
    );
    return $MockedSite;
}

#----------------------------------------------------------------------------------------
sub _getPaymentTypeMethod {
    my $self = shift;
    my ($isNameOrAlias) = @_;

    my $hAttributes = {
        'Alias' => 'MyPaymentTypeMethod',
        'NameOrAlias' => defined $isNameOrAlias ? 'MyPaymentTypeMethod' : undef
    };

    my $MockedPaymentTypeMethod = DE_EPAGES::Test::Mock::EpagesObject->new(
        'DE_EPAGES::ExternalPayment::API::PaymentTypeMethods', $hAttributes
    );
    $MockedPaymentTypeMethod->mock('newPaymentMethod', sub {
        return $self->_getPaymentMethodSpreedly();
    });
    return $MockedPaymentTypeMethod;
}

__PACKAGE__->runScript();